#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools

from .core import Backend
from ..expr.expressions import Summary
from ..expr.reduction import SequenceReduction
from ..expr.window import *  # don't remove
from ..expr.merge import *
from ..utils import traverse_until_source


class BaseRewriter(Backend):
    def __init__(self, expr_dag, traversed=None):
        self._dag = expr_dag
        self._indexer = itertools.count(0)
        self._traversed = traversed or set()

    def rewrite(self):
        for node in self._iter():
            self._traversed.add(id(node))
            self._visit_node(node)

        return self._dag.root

    def _iter(self):
        for node in traverse_until_source(self._dag, top_down=True,
                                          traversed=self._traversed):
            yield node

        while True:
            all_traversed = True
            for node in traverse_until_source(self._dag, top_down=True):
                if id(node) not in self._traversed:
                    all_traversed = False
                    yield node
            if all_traversed:
                break

    def _visit_node(self, node):
        try:
            node.accept(self)
        except NotImplementedError:
            return

    def _sub(self, expr, to_sub, parents=None):
        self._dag.substitute(expr, to_sub, parents=parents)

    def _parents(self, expr):
        return self._dag.successors(expr)

    def _rewrite_reduction_in_projection(self, expr):
        # FIXME how to handle nested reduction?
        if isinstance(expr, Summary):
            return

        collection = expr.input

        sink_selects = []
        columns = set()
        to_replace = []

        windows_rewrite = False
        for field in expr.fields:
            has_window = False
            traversed = set()
            for path in field.all_path(collection, strict=True):
                for node in path:
                    if id(node) in traversed:
                        continue
                    else:
                        traversed.add(id(node))
                    if isinstance(node, SequenceReduction) and not node._need_cache:
                        windows_rewrite = True
                        has_window = True

                        win = self._reduction_to_window(node)
                        window_name = '%s_%s' % (win.name, next(self._indexer))
                        sink_selects.append(win.rename(window_name))
                        to_replace.append((node, window_name))
                        break
                    elif isinstance(node, Column):
                        if node.input is not collection:
                            continue

                        if node.name in columns:
                            to_replace.append((node, node.name))
                            continue
                        columns.add(node.name)
                        select_field = collection[node.source_name]
                        if node.is_renamed():
                            select_field = select_field.rename(node.name)
                        sink_selects.append(select_field)
                        to_replace.append((node, node.name))

            if has_window:
                field._name = field.name

        if not windows_rewrite:
            return

        get = lambda x: x.name if not isinstance(x, six.string_types) else x
        projected = collection[sorted(sink_selects, key=get)]
        projected.optimize_banned = True  # TO prevent from optimizing
        expr.substitute(collection, projected, dag=self._dag)

        for col, col_name in to_replace:
            self._sub(col, projected[col_name].rename(col.name))

    def _rewrite_reduction_in_filter(self, expr):
        # FIXME how to handle nested reduction?
        collection = expr.input

        sink_selects = []
        columns = set()
        to_replace = []

        windows_rewrite = False
        traversed = set()
        for path in expr.predicate.all_path(collection, strict=True):
            for node in path:
                if id(node) in traversed:
                    continue
                else:
                    traversed.add(id(node))
                if isinstance(node, SequenceReduction):
                    windows_rewrite = True

                    win = self._reduction_to_window(node)
                    window_name = '%s_%s' % (win.name, next(self._indexer))
                    sink_selects.append(win.rename(window_name))
                    to_replace.append((node, window_name))
                    break
                elif isinstance(node, Column):
                    if node.input is not collection:
                        continue
                    if node.name in columns:
                        to_replace.append((node, node.name))
                        continue
                    columns.add(node.name)
                    select_field = collection[node.source_name]
                    if node.is_renamed():
                        select_field = select_field.rename(node.name)
                    sink_selects.append(select_field)
                    to_replace.append((node, node.name))

        for column_name in expr.schema.names:
            if column_name in columns:
                continue
            columns.add(column_name)
            sink_selects.append(column_name)

        if not windows_rewrite:
            return

        get = lambda x: x.name if not isinstance(x, six.string_types) else x
        projected = collection[sorted(sink_selects, key=get)]
        projected.optimize_banned = True  # TO prevent from optimizing
        expr.substitute(collection, projected, dag=self._dag)

        for col, col_name in to_replace:
            self._sub(col, projected[col_name])

        to_sub = expr[expr.schema.names]
        self._sub(expr, to_sub)

    def _reduction_to_window(self, expr):
        clazz = 'Cum' + expr.node_name
        return globals()[clazz](_input=expr.input, _data_type=expr.dtype)



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np
try:
    from scipy import sparse as sps
except ImportError:
    sps = None

from ..checksum_c cimport Checksum
from .block_encoder_c cimport Encoder
from . import errno
from ..pb import wire_format
from ..wireconstants import ProtoWireConstants
from ... import types, options

try:
    import pandas as pd
except ImportError:
    pd = None

cdef:
    uint32_t WIRETYPE_VARINT = wire_format.WIRETYPE_VARINT
    uint32_t WIRETYPE_FIXED32 = wire_format.WIRETYPE_FIXED32
    uint32_t WIRETYPE_FIXED64 = wire_format.WIRETYPE_FIXED64
    uint32_t WIRETYPE_LENGTH_DELIMITED = wire_format.WIRETYPE_LENGTH_DELIMITED

    uint32_t WIRE_TUNNEL_META_COUNT = ProtoWireConstants.TUNNEL_META_COUNT
    uint32_t WIRE_TUNNEL_META_CHECKSUM = ProtoWireConstants.TUNNEL_META_CHECKSUM
    uint32_t WIRE_TUNNEL_END_RECORD = ProtoWireConstants.TUNNEL_END_RECORD

    int BD_SUCCESS = errno.BD_SUCCESS
    int BD_BUFFER_EXHAUSTED = errno.BD_BUFFER_EXHAUSTED
    int BD_CHECKSUM_INVALID = errno.BD_CHECKSUM_INVALID
    int BD_COUNT_NOT_MATCH = errno.BD_COUNT_NOT_MATCH
    int BD_INVALID_STREAM_DATA = errno.BD_INVALID_STREAM_DATA
    int BD_INVALID_PB_TAG = errno.BD_INVALID_PB_TAG


cdef class BasePandasWriter:
    def __init__(self, object schema, object output_stream, columns=None):
        self._schema = schema
        if columns is None:
            self._columns = self._schema.columns
        else:
            self._columns = [self._schema[c] for c in columns]
        self._n_columns = len(self._columns)

        self._mem_cache_size = options.tunnel.pd_mem_cache_size

        self._stream = output_stream

        self._count = 0
        self._row_pos = 0
        self._crc = Checksum()
        self._crccrc = Checksum()

    cdef void _write_long_val(self, long val) nogil:
        self._crc.c_update_long(val)
        self._encoder.append_sint32(val)

    cdef void _write_long(self, ArrayVariantPtrs &aptr, int index) nogil:
        cdef long val = aptr.v.pl_array[index]
        self._crc.c_update_long(val)
        self._encoder.append_sint32(val)

    cdef void _write_bool(self, ArrayVariantPtrs &aptr, int index) nogil:
        cdef np.npy_bool val = aptr.v.pbool_array[index]
        self._crc.c_update_bool(val)
        self._encoder.append_bool(val)

    cdef void _write_float(self, ArrayVariantPtrs &aptr, int index) nogil:
        cdef float val = aptr.v.pflt_array[index]
        self._crc.c_update_float(val)
        self._encoder.append_float(val)

    cdef void _write_double(self, ArrayVariantPtrs &aptr, int index) nogil:
        cdef double val = aptr.v.pdbl_array[index]
        self._crc.c_update_double(val)
        self._encoder.append_double(val)

    cpdef reset_positions(self):
        self._row_pos = 0
        self._encoder.init(self._mem_cache, self._mem_cache_size)

    cpdef init_cache(self):
        self._mem_cache = bytearray(self._mem_cache_size)
        self._mem_cache_view = memoryview(self._mem_cache)
        self._encoder = Encoder()
        self.reset_positions()

    cpdef write_stream(self, object data, int length):
        self._stream.write(self._mem_cache_view[:length])

    cdef int _write_single_ndarray_nogil(self, ArrayVariantPtrs &col_ptr, vector[int] &dims,
                                         vector[int] &col_to_dim, long start_pos, long limit,
                                         vector[long] &dim_offsets) nogil:
        cdef:
            int i, j, dim_id
            long flat_pos
            long max_pos = 1
            long rest_pos
            vector[int] array_pos

        array_pos.resize(dims.size())
        rest_pos = start_pos
        for i in reversed(range(dims.size())):
            max_pos *= dims[i]
            array_pos[i] = rest_pos % dims[i]
            rest_pos /= dims[i]

        if limit > 0:
            max_pos = min(limit, max_pos)

        self._crc.c_setvalue(0)
        for i in range(start_pos, max_pos):
            self._crc.c_setvalue(0)

            for j in range(col_to_dim.size()):
                if col_to_dim[j] < 0:
                    continue
                self._crc.c_update_int(j + 1)
                if col_to_dim[j] == 0:
                    self._encoder.append_tag(j + 1, col_ptr.wire_type)
                    if self._encoder.get_last_error() != BD_SUCCESS:
                        return i - start_pos

                    self._nogil_writers[0](self, col_ptr, i)
                    if self._encoder.get_last_error() != BD_SUCCESS:
                        return i - start_pos
                else:
                    dim_id = col_to_dim[j] - 1
                    self._encoder.append_tag(j + 1, WIRETYPE_VARINT)
                    if self._encoder.get_last_error() != BD_SUCCESS:
                        return i - start_pos

                    self._write_long_val(array_pos[dim_id] + dim_offsets[dim_id])
                    if self._encoder.get_last_error() != BD_SUCCESS:
                        return i - start_pos

            checksum = <int32_t>self._crc.c_getvalue()

            self._encoder.append_tag(WIRE_TUNNEL_END_RECORD, WIRETYPE_VARINT)
            if self._encoder.get_last_error() != BD_SUCCESS:
                return i - start_pos
            self._encoder.append_uint32(checksum)
            if self._encoder.get_last_error() != BD_SUCCESS:
                return i - start_pos
            self._crccrc.c_update_int(checksum)

            self._row_pos = self._encoder.position()

            array_pos[array_pos.size() - 1] += 1
            for j in reversed(range(1, array_pos.size())):
                if array_pos[j] >= dims[j]:
                    array_pos[j - 1] += 1
                    array_pos[j] = 0
                else:
                    break

        return max_pos - start_pos

    cdef int _write_dims_nogil(self, vector[ArrayVariantPtrs] &col_ptrs, vector[int] &col_to_dim,
                               long start_row, long limit) nogil:
        cdef:
            int i
            int dim_id
            int row_id

        for row_id in range(start_row, limit):
            self._crc.c_setvalue(0)

            for i in range(col_to_dim.size()):
                if col_to_dim[i] < 0:
                    continue
                self._crc.c_update_int(i + 1)
                dim_id = col_to_dim[i] - 1

                self._encoder.append_tag(i + 1, col_ptrs[dim_id].wire_type)
                if self._encoder.get_last_error() != BD_SUCCESS:
                    return row_id - start_row

                self._nogil_writers[dim_id](self, col_ptrs[dim_id], row_id)
                if self._encoder.get_last_error() != BD_SUCCESS:
                    return row_id - start_row

            checksum = <int32_t>self._crc.c_getvalue()

            self._encoder.append_tag(WIRE_TUNNEL_END_RECORD, WIRETYPE_VARINT)
            if self._encoder.get_last_error() != BD_SUCCESS:
                return row_id - start_row
            self._encoder.append_uint32(checksum)
            if self._encoder.get_last_error() != BD_SUCCESS:
                return row_id - start_row
            self._crccrc.c_update_int(checksum)

            self._row_pos = self._encoder.position()

        return limit - start_row

    cpdef _write_single_array(self, object data, object columns, long limit, object dim_offsets):
        cdef:
            int i
            long start_pos
            long count_delta
            long total_size
            vector[int] dims
            vector[long] dim_offset_vct
            vector[int] col_to_dim

            ArrayVariantPtrs col_ptr
            Py_buffer buf

            dict col_ids
            object array_type
            object val_column

        total_size = 1
        dims.resize(len(data.shape))
        for i in range(dims.size()):
            dims[i] = data.shape[i]
            total_size *= dims[i]

        dim_offset_vct.resize(len(data.shape))
        if dim_offsets is None:
            for i in range(dim_offset_vct.size()):
                dim_offset_vct[i] = 0
        else:
            for i in range(dim_offset_vct.size()):
                dim_offset_vct[i] = dim_offsets[i]

        col_to_dim.resize(data.ndim + 1)
        if columns is None:
            if self._n_columns != col_to_dim.size():
                raise ValueError('Column number not consistent with array shape: num of '
                                 'columns should be 1 + ndim')
            for i in range(col_to_dim.size()):
                col_to_dim[i] = i + 1
            col_to_dim[col_to_dim.size() - 1] = 0
            val_column = self._schema[-1]
        else:
            col_ids = dict()
            for idx, col_name in enumerate(columns):
                col_ids[col_name] = idx + 1
            col_ids[columns[-1]] = 0

            for idx, col in enumerate(self._schema):
                if col.name in col_ids:
                    i = idx
                    col_to_dim[i] = col_ids[col.name]
                    if col_to_dim[i] == 0:
                        val_column = col
                else:
                    col_to_dim[i] = -1

        self._nogil_writers.resize(1)
        if val_column.type == types.float_:
            col_ptr.wire_type = WIRETYPE_FIXED32
            data = data.astype(np.float_) if data.dtype != np.float_ else data
            col_ptr.v.pflt_array = <float *>np.PyArray_DATA(data)
            self._nogil_writers[0] = self._write_float
        elif val_column.type == types.double:
            col_ptr.wire_type = WIRETYPE_FIXED64
            data = data.astype(np.double) if data.dtype != np.double else data
            col_ptr.v.pdbl_array = <double *>np.PyArray_DATA(data)
            self._nogil_writers[0] = self._write_double
        elif val_column.type in types.integer_types:
            col_ptr.wire_type = WIRETYPE_VARINT
            data = data.astype(np.int64) if data.dtype != np.int64 else data
            col_ptr.v.pl_array = <int64_t *>np.PyArray_DATA(data)
            self._nogil_writers[0] = self._write_long
        elif val_column.type == types.boolean:
            col_ptr.wire_type = WIRETYPE_VARINT
            data = data.astype(np.bool_) if data.dtype != np.bool_ else data
            col_ptr.v.pbool_array = <np.npy_bool *>np.PyArray_DATA(data)
            self._nogil_writers[0] = self._write_bool

        if limit <= 0:
            limit = total_size

        start_pos = 0
        self.init_cache()
        while start_pos < limit:
            self.reset_positions()
            count_delta = self._write_single_ndarray_nogil(
                col_ptr, dims, col_to_dim, start_pos, limit, dim_offset_vct)
            self._count += count_delta
            start_pos += count_delta
            self.write_stream(self._mem_cache, self._row_pos)

    cpdef _write_dims(self, object data, object columns, long limit):
        cdef:
            int i
            long start_pos
            long count_delta
            long total_size
            vector[int] dims
            vector[int] col_to_dim

            vector[ArrayVariantPtrs] col_ptrs
            dict col_dict
            dict col_idx

            object col_data
            int64_t[:] int_mmap
            float[:] flt_mmap
            double[:] dbl_mmap
            np.ndarray[np.npy_bool, ndim=1, cast=True] bool_array

        col_dict = dict()
        col_idx = dict()

        if isinstance(data, dict):
            col_dict = data
        elif pd and isinstance(data, pd.DataFrame):
            for col_name in data:
                col_dict[col_name] = data[col_name].values
        elif columns is not None:
            for col_name, buf in zip(columns, data):
                col_dict[col_name] = buf
        else:
            for col, buf in zip(self._columns, data):
                col_dict[col.name] = buf

        if limit < 0:
            limit = 0x7fffffff
            for buf in col_dict.values():
                limit = min(limit, len(buf))

        col_ptrs.resize(self._n_columns)
        col_to_dim.resize(self._n_columns)

        i = 0
        self._nogil_writers.resize(len(col_dict))
        for col_name, col_data in col_dict.items():
            data_type = self._schema[col_name].type
            if data_type == types.float_:
                col_ptrs[i].wire_type = WIRETYPE_FIXED32
                flt_mmap = col_data.astype(np.float_) if col_data.dtype != np.float_ else col_data
                col_ptrs[i].v.pflt_array = &flt_mmap[0]
                self._nogil_writers[i] = self._write_float
            elif data_type == types.double:
                col_ptrs[i].wire_type = WIRETYPE_FIXED64
                dbl_mmap = col_data.astype(np.double) if col_data.dtype != np.double else col_data
                col_ptrs[i].v.pdbl_array = &dbl_mmap[0]
                self._nogil_writers[i] = self._write_double
            elif data_type in types.integer_types:
                col_ptrs[i].wire_type = WIRETYPE_VARINT
                int_mmap  = col_data.astype(np.int64) if col_data.dtype != np.int64 else col_data
                col_ptrs[i].v.pl_array = &int_mmap[0]
                self._nogil_writers[i] = self._write_long
            elif data_type == types.boolean:
                col_ptrs[i].wire_type = WIRETYPE_VARINT
                bool_array  = col_data.astype(np.bool_) if col_data.dtype != np.bool_ else col_data
                col_ptrs[i].v.pbool_array = <np.npy_bool *>bool_array.data
                self._nogil_writers[i] = self._write_bool
            col_idx[col_name] = i
            i += 1

        for i in range(self._n_columns):
            col_name = self._columns[i].name

            if col_name not in col_idx:
                col_to_dim[i] = -1
                continue
            else:
                col_to_dim[i] = col_idx[col_name] + 1

        start_pos = 0
        self.init_cache()
        while start_pos < limit:
            self.reset_positions()
            count_delta = self._write_dims_nogil(col_ptrs, col_to_dim, start_pos, limit)
            self._count += count_delta
            start_pos += count_delta
            self._stream.write(self._mem_cache_view[:self._row_pos])

    cpdef write(self, object data, object columns=None, long limit=-1, object dim_offsets=None):
        """
        Write a numpy array, a pandas DataFrame or a dict of column names and columns into a table.
        When writing a numpy array, the indices and value of every element is written. The indices
        are written before values. The argument ``columns`` determines the subset and order of data.
        
        Currently only ``bigint``, ``float``, ``double`` and ``boolean`` are supported.
        
        :param data: data to write, can be numpy array, dict or pandas DataFrame
        :param columns: column names to write
        :param limit: total number of records to write
        :param dim_offsets: offsets for every dimensions, only applicable for arrays
        :return: number of records written
        """
        if isinstance(data, np.ndarray):
            return self._write_single_array(data, columns, limit, dim_offsets)

        if sps:
            if isinstance(data, sps.csr_matrix):
                data = data.tocoo()
            if isinstance(data, sps.coo_matrix):
                row = data.row
                col = data.col
                data_col = data.data
                del data
                if dim_offsets is not None:
                    row += dim_offsets[0]
                    col += dim_offsets[1]
                return self._write_dims([row, col, data_col], columns, limit)

        return self._write_dims(data, columns, limit)

    cpdef close(self):
        self.reset_positions()
        self._encoder.append_tag(WIRE_TUNNEL_META_COUNT, WIRETYPE_VARINT)
        self._encoder.append_sint64(self._count)
        self._encoder.append_tag(WIRE_TUNNEL_META_CHECKSUM, WIRETYPE_VARINT)
        self._encoder.append_uint32(<uint32_t>self._crccrc.getvalue())
        self._stream.write(self._mem_cache_view[:self._encoder.position()])
        self.flush()

    cpdef flush(self):
        self._stream.flush()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # if an error occurs inside the with block, we do not commit
        if exc_val is not None:
            return
        self.close()



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging

from odps import (core, accounts)


CONF_FILENAME = '.pyouconfig'

logger = logging.getLogger(__name__)
_odps = None


class UDFToolError(Exception):
    pass


def require_conf(func):
    def f(*args, **kwargs):
        if _odps is None:
            get_conf()
        return func(*args, **kwargs)
    return f
    

def get_conf():
    import ConfigParser
    global _odps

    home_dir = os.environ.get('HOME')
    if not home_dir:
        raise UDFToolError('Cannot find home dir, '
                           'perhaps you are using windowns :(')
    f = os.path.join(home_dir, CONF_FILENAME)
    if not os.path.exists(f):
        access_id = raw_input('access_id:')
        
    config = ConfigParser.RawConfigParser()
    config.read(f)
    access_id = config.get('access_id')
    access_key = config.get('secret_access_key')
    end_point = config.get('endpoint')
    project = config.get('project')
    _odps = core.ODPS(access_id, access_key,
                      project, end_point)


@require_conf
def get_cache_table(name):
    return _odps.read_table(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import exporters
from .core import AlgoCollectionExpr
from ..utils import ML_ARG_PREFIX


_MERGE_DATA_ARGS = """
inputTableNames outputTableName selectedColNamesList inputPartitionsInfoList outputPartition
autoRenameCol
""".split()
_MERGE_DATA_EXPORTERS = {
    "outputTableName": lambda e: exporters.get_output_table_name(e, "output"),
    "inputTableNames": lambda e: e._export_input_table_names(),
    "inputPartitionsInfoList": lambda e: e._export_input_table_partitions(),
    "selectedColNamesList": lambda e: e._export_selected_cols(),
}
_MERGE_DATA_META = {
    'xflowName': 'AppendColumns',
    'xflowProjectName': 'algo_public',
}


class BaseMergeDataCollectionExpr(AlgoCollectionExpr):
    __slots__ = '_selected_merge_cols', '_excluded_merge_cols'
    node_name = 'MergeData'
    _algo = 'MergeData'
    _exported = _MERGE_DATA_ARGS
    _exporters = _MERGE_DATA_EXPORTERS
    algo_meta = _MERGE_DATA_META

    def _init(self, *args, **kwargs):
        self._init_attr('_selected_merge_cols', kwargs.pop('selected_cols', dict()))
        self._init_attr('_excluded_merge_cols', kwargs.pop('excluded_cols', dict()))

        super(BaseMergeDataCollectionExpr, self)._init(*args, **kwargs)

        self._init_attr('_params', dict())
        self._init_attr('_engine_kw', dict())

    def _iter_inputs(self):
        for name, df in self.iter_args():
            if name.startswith(ML_ARG_PREFIX):
                yield name[len(ML_ARG_PREFIX):], df

    def _export_input_table_names(self):
        tables = []
        for name, expr in self._iter_inputs():
            tn = exporters.get_input_table_name(self, name)
            if expr is not None and not tn:
                return None
            tables.append(tn if tn else '')
        return ','.join(tables)

    def _export_input_table_partitions(self):
        parts = []
        for name, _ in self._iter_inputs():
            tn = exporters.get_input_table_partitions(self, name)
            parts.append(tn if tn else '')
        if all(not tn for tn in parts):
            return None
        return ','.join(parts)

    def _export_selected_cols(self):
        def fetch_ds_cols(seq, df):
            if seq in self._selected_merge_cols:
                return self._selected_merge_cols[seq]
            elif seq in self._excluded_merge_cols:
                return [f for f in df._ml_fields if f.role and f.name not in self._excluded_merge_cols[seq]]
            else:
                return [f for f in df._ml_fields if f.role]

        return ';'.join(','.join(f.name for f in fetch_ds_cols(seq, df_tuple[1]))
                        for seq, df_tuple in enumerate(self._iter_inputs()))


def build_merge_expr(num_inputs):
    inputs = [ML_ARG_PREFIX + 'input%d' % (idx + 1) for idx in range(num_inputs)]

    class MergeDataCollectionExpr(BaseMergeDataCollectionExpr):
        _args = tuple(inputs)

    return MergeDataCollectionExpr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import types as odps_types

try:
    import pyarrow as pa
except (AttributeError, ImportError):
    pa = None


if pa is not None:
    _ODPS_ARROW_TYPE_MAPPING = {
        odps_types.string: pa.string(),
        odps_types.binary: pa.binary(),
        odps_types.tinyint: pa.int8(),
        odps_types.smallint: pa.int16(),
        odps_types.int_: pa.int32(),
        odps_types.bigint: pa.int64(),
        odps_types.boolean: pa.bool_(),
        odps_types.float_: pa.float32(),
        odps_types.double: pa.float64(),
        odps_types.date: pa.date32(),
        odps_types.datetime: pa.timestamp('ms'),
        odps_types.timestamp: pa.timestamp('ns'),
    }
else:
    _ODPS_ARROW_TYPE_MAPPING = {}


def odps_type_to_arrow_type(odps_type):
    from ... import types

    if odps_type in _ODPS_ARROW_TYPE_MAPPING:
        col_type = _ODPS_ARROW_TYPE_MAPPING[odps_type]
    else:
        if isinstance(odps_type, types.Array):
            col_type = pa.list_(odps_type_to_arrow_type(odps_type.value_type))
        elif isinstance(odps_type, types.Map):
            col_type = pa.map_(
                odps_type_to_arrow_type(odps_type.key_type),
                odps_type_to_arrow_type(odps_type.value_type),
            )
        elif isinstance(odps_type, types.Decimal):
            precision = odps_type.precision or types.Decimal._max_precision
            scale = odps_type.scale or types.Decimal._max_scale
            if odps_type.precision is None and not hasattr(pa, "decimal256"):
                # need to be less than minimal allowed digits of pa.decimal128
                precision = min(precision, 38)
            decimal_cls = getattr(pa, "decimal256") if precision > 38 else pa.decimal128
            col_type = decimal_cls(precision, scale)
        elif isinstance(odps_type, types.Struct):
            fields = [
                (k, odps_type_to_arrow_type(v))
                for k, v in odps_type.field_types.items()
            ]
            col_type = pa.struct(fields)
        elif isinstance(odps_type, odps_types.IntervalDayTime):
            col_type = pa.struct([("sec", pa.int64()), ("nano", pa.int32())])
        else:
            raise TypeError('Unsupported type: {}'.format(odps_type))
    return col_type


def odps_schema_to_arrow_schema(odps_schema):

    arrow_schema = []
    for schema in odps_schema.simple_columns:
        col_name = schema.name
        col_type = odps_type_to_arrow_type(schema.type)
        arrow_schema.append(pa.field(col_name, col_type))

    return pa.schema(arrow_schema)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from ..core import Backend
from ...expr.core import ExprDictionary
from ...expr.expressions import *
from ...expr.groupby import GroupByCollectionExpr
from ...expr.reduction import SequenceReduction, GroupedSequenceReduction
from ...expr.merge import JoinCollectionExpr
from ...expr.window import Window
from ...expr.utils import select_fields
from ...utils import traverse_until_source
from .... import utils
from .columnpruning import ColumnPruning
from .predicatepushdown import PredicatePushdown
from .utils import change_input


class Optimizer(Backend):
    def __init__(self, dag):
        self._dag = dag

    def optimize(self):
        if options.df.optimize:
            if options.df.optimizes.cp:
                ColumnPruning(self._dag).prune()
            if options.df.optimizes.pp:
                PredicatePushdown(self._dag).pushdown()

        for node in traverse_until_source(self._dag, top_down=True):
            try:
                node.accept(self)
            except NotImplementedError:
                continue

        # from down up do again
        for node in traverse_until_source(self._dag):
            try:
                node.accept(self)
            except NotImplementedError:
                continue

        return self._dag.root

    def _sub(self, expr, to_sub, parents=None):
        self._dag.substitute(expr, to_sub, parents=parents)

    def visit_filter_collection(self, expr):
        if not options.df.optimize:
            return

        if isinstance(expr.input, GroupByCollectionExpr) and \
                not expr.input.optimize_banned:
            # move filter on GroupBy to GroupBy's having
            grouped = expr.input
            predicate = self._do_compact(expr.predicate, expr.input)
            if predicate is None:
                predicate = expr.predicate

            having = grouped.having
            if having is not None:
                predicates = having & predicate
            else:
                predicates = predicate

            grouped._having = predicates
            self._sub(expr, grouped)
        elif isinstance(expr.input, FilterCollectionExpr):
            filters = [expr]
            node = expr.input
            while isinstance(node, FilterCollectionExpr):
                filters.append(node)
                node = node.input
            self._compact_filters(*filters)

        raise NotImplementedError

    @classmethod
    def get_compact_filters(cls, dag, *filters):
        input = filters[-1].input

        get_field = lambda n, col: input[col]
        for filter in filters:
            change_input(filter, filter.input, input, get_field, dag)

        predicate = reduce(operator.and_, [f.predicate for f in filters[::-1]])
        return FilterCollectionExpr(input, predicate, _schema=input.schema)

    def _compact_filters(self, *filters):
        new_filter = self.get_compact_filters(self._dag, *filters)
        self._sub(filters[0], new_filter)

    def visit_project_collection(self, expr):
        # Summary does not attend here
        if not options.df.optimize:
            return

        compacted = self._visit_need_compact_collection(expr)
        if compacted:
            expr = compacted

        if isinstance(expr, ProjectCollectionExpr) and \
                isinstance(expr.input, GroupByCollectionExpr) and \
                not expr.input.optimize_banned:
            # compact projection into Groupby
            grouped = expr.input

            selects = []

            for n in expr.traverse(top_down=True, unique=True,
                                   stop_cond=lambda x: x is grouped):
                # stop compact
                if isinstance(n, (Window, SequenceReduction)):
                    return

            for field in expr._fields:
                selects.append(self._do_compact(field, grouped) or field)

            grouped._aggregations = grouped._fields = selects
            grouped._schema = TableSchema.from_lists([f.name for f in selects],
                                                [f.dtype for f in selects])
            self._sub(expr, grouped)

            return

    def visit_groupby(self, expr):
        if not options.df.optimize:
            return

        # we do not do compact on the projections from Join
        input = expr.input
        while isinstance(input, ProjectCollectionExpr):
            input = input._input
        if isinstance(input, JoinCollectionExpr):
            return

        if len(expr._aggregations) == 1 and \
                isinstance(expr._aggregations[0], GroupedSequenceReduction) and \
                isinstance(expr._aggregations[0]._input, CollectionExpr):
            # we just skip the case: df.groupby(***).count()
            return

        self._visit_need_compact_collection(expr)

    def visit_distinct(self, expr):
        if not options.df.optimize:
            return
        self._visit_need_compact_collection(expr)

    def visit_apply_collection(self, expr):
        if not options.df.optimize:
            return
        if expr._lateral_view:
            return
        if (
            isinstance(expr.input, JoinCollectionExpr)
            and (expr.input._mapjoin or expr.input._skewjoin)
        ):
            return
        self._visit_need_compact_collection(expr)

    def _visit_need_compact_collection(self, expr):
        compacted = self._compact(expr)
        if compacted is None:
            return

        if expr is not compacted:
            self._sub(expr, compacted)
            return compacted

    def _compact(self, expr):
        to_compact = [expr, ]

        for node in traverse_until_source(expr, top_down=True, unique=True):
            if node is expr:
                continue
            if not isinstance(node, CollectionExpr):
                continue

            # We do not handle collection with Scalar column or window function here
            # TODO think way to compact in this situation
            if isinstance(node, ProjectCollectionExpr) and \
                    not node.optimize_banned and \
                    not any(isinstance(n, Window) for n in node._fields):
                valid = True
                for it in itertools.chain(*(node.all_path(to_compact[-1]))):
                    if isinstance(it, CollectionExpr) and \
                            any(isinstance(n.input, LateralViewCollectionExpr) for n in it.columns):
                        valid = False
                        break
                    if isinstance(it, SequenceReduction):
                        valid = False
                        break
                if not valid:
                    break
                to_compact.append(node)
            else:
                break

        if len(to_compact) <= 1:
            return

        return self._do_compact(expr, *to_compact[1:][::-1])

    def _do_compact(self, expr, *to_compact):
        retval = expr

        collection_dict = ExprDictionary()
        for coll in to_compact:
            collection_dict[coll] = True

        for node in expr.traverse(top_down=True, unique=True,
                                  stop_cond=lambda x: x is to_compact[-1]):
            if node in collection_dict:
                parents = self._dag.successors(node)
                for parent in parents:
                    if isinstance(parent, Column):
                        col = parent

                        col_name = col.source_name or col.name
                        field = self._get_field(node, col_name)
                        if col.is_renamed():
                            field = field.rename(col.name)
                        else:
                            field = field.copy()

                        self._sub(col, field)

                        if col is retval:
                            retval = field
                    else:
                        parent.substitute(node, node.input, dag=self._dag)

        return retval

    def _get_fields(self, collection):
        fields = select_fields(collection)
        if isinstance(collection, GroupByCollectionExpr) and \
                collection._having is not None:
            # add GroupbyCollectionExpr.having to broadcast fields
            fields.append(collection._having)
        return fields

    def _get_field(self, collection, name):
        # FIXME: consider name with upper letters
        name = utils.to_str(name)

        if isinstance(collection, GroupByCollectionExpr):
            return collection._name_to_exprs()[name]

        name_idxes = collection.schema._name_indexes
        if name.lower() in name_idxes:
            idx = name_idxes[name.lower()]
        else:
            idx = name_idxes[name]
        return self._get_fields(collection)[idx]


# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from ..expr.exporters import get_enable_sparse, get_input_field_names
from ..expr.mixin import ml_collection_mixin
from ...compat import Enum


class RecommendFieldRole(Enum):
    REC_USER_ID = 'USER_ID'
    REC_ITEM = 'REC_ITEM'
    REC_SEQUENCE = 'REC_SEQUENCE'
    REC_PAYLOAD = 'REC_PAYLOAD'


@ml_collection_mixin
class RecommendMLMixIn(object):
    __slots__ = ()

    field_role_enum = RecommendFieldRole
    non_feature_roles = set([RecommendFieldRole.REC_ITEM, ])


"""
Common recommend exporters
"""
get_rec_user_id_column = partial(get_input_field_names, field_role=RecommendFieldRole.REC_USER_ID)
get_rec_item_column = partial(get_input_field_names, field_role=RecommendFieldRole.REC_ITEM)
get_rec_sequence_column = partial(get_input_field_names, field_role=RecommendFieldRole.REC_SEQUENCE)
get_rec_payload_column = partial(get_input_field_names, field_role=RecommendFieldRole.REC_PAYLOAD)


def get_etrec_table_format(expr, param_name, input_name):
    if expr._params.get(param_name, None):
        return expr._params[param_name]
    return 'user-item' if not get_enable_sparse(expr, param_name, input_name) else 'items'


def get_rec_triple_selected_col_names(expr, param_name, input_name):
    if expr._params.get(param_name, None):
        return expr._params[param_name]
    cols = [get_rec_user_id_column(expr, param_name, input_name)[0],
            get_rec_item_column(expr, param_name, input_name)[0]]
    payload_col = get_rec_payload_column(expr, param_name, input_name)[0]
    if payload_col:
        cols.append(payload_col)
    return ','.join(cols)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from datetime import timedelta

from ..core import Backend
from ..errors import CompileError
from ...expr.reduction import *
from ...expr.arithmetic import *
from ...expr.datetimes import *
from ...expr.window import *
from ...expr.merge import *
from ...expr.utils import highest_precedence_data_type
from ... import types as df_types
from ...utils import is_constant_scalar, traverse_until_source, is_source_collection
from ....compat import lzip
from .... import utils
from . import types

try:
    from sqlalchemy import Table as SATable, select, func, case, \
        extract, desc, distinct, literal, join, union, union_all
    from sqlalchemy.sql.expression import Alias, text

    # define the compile ext
    from .ext import *

    has_sqlalchemy = True
except ImportError:
    has_sqlalchemy = False


BINARY_OP = {
    'Add': operator.add,
    'Substract': operator.sub,
    'Multiply': operator.mul,
    'Divide': operator.div if six.PY2 else operator.truediv,
    'Mod': operator.mod,
    'FloorDivide': operator.floordiv,
    'Greater': operator.gt,
    'GreaterEqual': operator.ge,
    'Less': operator.lt,
    'LessEqual': operator.le,
    'Equal': operator.eq,
    'NotEqual': operator.ne,
    'And': operator.and_,
    'Or': operator.or_
}

UNARY_OP = {
    'Negate': operator.neg,
    'Invert': operator.inv,
}

DATE_KEY_DIC = {
    'Day': 'days',
    'Hour': 'hours',
    'Minute': 'minutes',
    'Second': 'seconds',
    'MilliSecond': 'milliseconds',
    'MicroSecond': 'microseconds',
}

MATH_COMPILE_DIC = {
    'Abs': 'abs',
    'Sqrt': 'sqrt',
    'Sin': 'sin',
    'Cos': 'cos',
    'Tan': 'tan',
    'Exp': 'exp',
    'Arccos': 'acos',
    'Arcsin': 'asin',
    'Arctan': 'atan',
    'Ceil': 'ceil',
    'Floor': 'floor'
}

WINDOW_COMPILE_DIC = {
    'CumSum': 'sum',
    'CumMean': 'avg',
    'CumStd': 'stddev',
    'CumMax': 'max',
    'CumMin': 'min',
    'CumCount': 'count',
    'Lag': 'lag',
    'Lead': 'lead',
    'Rank': 'rank',
    'DenseRank': 'dense_rank',
    'PercentRank': 'percent_rank',
    'RowNumber': 'row_number'
}

DATE_PARTS_DIC = {
    'Year': 'year',
    'Month': 'month',
    'Day': 'day',
    'Hour': 'hour',
    'Minute': 'minute',
    'Second': 'second',
    'WeekOfYear': 'week',
    'DayOfYear': 'doy',
    'MicroSecond': 'microseconds',
    'UnixTimestamp': 'epoch',
}


class SQLAlchemyCompiler(Backend):
    def __init__(self, expr_dag):
        self._expr_dag = expr_dag
        self._expr_to_sqlalchemy = ExprDictionary()
        self._id_gen = itertools.count(1)

        self._sa_engine = None

    def _new_alias(self):
        return 't%s' % next(self._id_gen)

    def compile(self, expr, traversed=None):
        if traversed is None:
            traversed = set()

        for node in traverse_until_source(expr):
            if id(node) not in traversed:
                node.accept(self)
                traversed.add(id(node))

        sa_expr = self._expr_to_sqlalchemy[self._expr_dag.root]
        if is_source_collection(expr):
            sa_expr = select([sa_expr])
        return sa_expr

    def _add(self, expr, op):
        self._expr_to_sqlalchemy[expr] = op

    def _gen_select_columns(self, fields):
        sa_exprs = []
        for field in fields:
            if not isinstance(field, Scalar) or field._value is None:
                sa_exprs.append(self._expr_to_sqlalchemy[field].label(field.name))
            else:
                sa_exprs.append(
                    literal(self._expr_to_sqlalchemy[field]).label(field.name))
        return sa_exprs

    def visit_source_collection(self, expr):
        table = next(expr.data_source())

        if not isinstance(table, SATable):
            raise ValueError('Source data must be a sqlalchemy table')

        if table.bind and self._sa_engine is None:
            self._sa_engine = table.bind

        self._add(expr, table.alias(self._new_alias()))

    def visit_project_collection(self, expr):
        selects = select(self._gen_select_columns(expr._fields))\
            .select_from(self._expr_to_sqlalchemy[expr.input])

        if expr is not self._expr_dag.root:
            selects = selects.alias(self._new_alias())
        self._add(expr, selects)

    def visit_apply_collection(self, expr):
        raise NotImplementedError

    def visit_filter_collection(self, expr):
        input = self._expr_to_sqlalchemy[expr.input]
        predicate = self._expr_to_sqlalchemy[expr._predicate]
        filtered = input.select(predicate)

        if expr is not self._expr_dag.root:
            filtered = filtered.alias(self._new_alias())
        self._add(expr, filtered)

    def visit_slice_collection(self, expr):
        input = self._expr_to_sqlalchemy[expr.input]
        sliced = expr._indexes
        if sliced[2] is not None:
            raise NotImplementedError
        if sliced[0] is not None and sliced[0].value < 0:
            raise CompileError('start number must be greater than 0')
        if sliced[1] is not None and sliced[1].value <= 0:
            raise CompileError('end number must be greater than 0')

        kw = dict()
        if sliced[0] is not None and sliced[0].value > 0:
            kw['offset'] = sliced[0].value
        if sliced[1] is not None and sliced[1].value > 0:
            kw['limit'] = sliced[1].value
        input = input.select(**kw)

        if expr is not self._expr_dag.root:
            input = input.alias(self._new_alias())

        self._add(expr, input)

    def visit_element_op(self, expr):
        input = self._expr_to_sqlalchemy.get(expr.input)
        if isinstance(expr, element.IsNull):
            sa_expr = input.is_(None)
        elif isinstance(expr, element.NotNull):
            sa_expr = input.isnot(None)
        elif isinstance(expr, element.FillNa):
            sa_expr = case([(input.is_(None), expr.fill_value)], else_=input)
        elif isinstance(expr, (element.IsIn, element.NotIn)):
            op = input.in_ if isinstance(expr, element.IsIn) else input.notin_
            if expr._values is None:
                sa_expr = op([None])
            elif len(expr._values) == 1 and isinstance(expr._values[0], SequenceExpr):
                right = select([self._expr_to_sqlalchemy[expr._values[0]]])
                sa_expr = op(right)
            else:
                sa_expr = op(tuple(self._expr_to_sqlalchemy[it] for it in expr._values))
        elif isinstance(expr, element.Between):
            if not expr.inclusive:
                raise NotImplementedError
            sa_expr = input.between(
                self._expr_to_sqlalchemy[expr._left],
                self._expr_to_sqlalchemy[expr._right]
            )
        elif isinstance(expr, element.IfElse):
            sa_expr = case([(input, self._expr_to_sqlalchemy[expr._then])],
                           else_=self._expr_to_sqlalchemy[expr._else])
        elif isinstance(expr, element.Switch):
            conditions = [self._expr_to_sqlalchemy[cond] for cond in expr._conditions]
            thens = [self._expr_to_sqlalchemy[then] for then in expr._thens]
            sa_else = self._expr_to_sqlalchemy[expr._default] \
                if expr._default is not None else expr._default
            if expr._input is None:
                sa_expr = case(lzip(conditions, thens), else_=sa_else)
            else:
                sa_expr = case(dict(lzip(conditions, thens)),
                               value=input, else_=sa_else)
        else:
            raise NotImplementedError

        self._add(expr, sa_expr)

    def visit_binary_op(self, expr):
        if isinstance(expr, Power):
            op = func.pow
        elif isinstance(expr, FloorDivide):
            op = operator.div if six.PY2 else operator.truediv
        elif isinstance(expr, (Add, Substract)) and expr.dtype == df_types.datetime:
            if isinstance(expr, Add) and \
                    all(child.dtype == df_types.datetime for child in (expr.lhs, expr.rhs)):
                raise CompileError('Cannot add two datetimes')
            if isinstance(expr.rhs, DTScalar) or (isinstance(expr, Add) and expr.lhs, DTScalar):
                if isinstance(expr.rhs, DTScalar):
                    dt, scalar = expr.lhs, expr.rhs
                else:
                    dt, scalar = expr.rhs, expr.lhs
                val = scalar.value
                if isinstance(expr, Substract):
                    val = -val

                dt_type = type(scalar).__name__[:-6]
                sa_dt = self._expr_to_sqlalchemy[dt]
                try:
                    key = DATE_KEY_DIC[dt_type]
                except KeyError:
                    raise NotImplementedError
                if self._sa_engine and self._sa_engine.name == 'mysql':
                    if dt_type == 'MilliSecond':
                        val, dt_type = val * 1000, 'MicroSecond'
                    sa_expr = func.date_add(sa_dt, text('interval %d %s' % (val, dt_type.lower())))
                else:
                    sa_expr = sa_dt + timedelta(**{key: val})
                self._add(expr, sa_expr)
                return
            else:
                raise NotImplementedError
        elif isinstance(expr, Substract) and expr._lhs.dtype == df_types.datetime and \
                expr._rhs.dtype == df_types.datetime:
            sa_expr = self._expr_to_sqlalchemy[expr._lhs] - self._expr_to_sqlalchemy[expr._rhs]
            if self._sa_engine and self._sa_engine.name == 'mysql':
                sa_expr = func.abs(func.microsecond(sa_expr)
                                   .cast(types.df_type_to_sqlalchemy_type(expr.dtype))) / 1000
            else:
                sa_expr = func.abs(extract('MICROSECONDS', sa_expr)
                                   .cast(types.df_type_to_sqlalchemy_type(expr.dtype))) / 1000
            self._add(expr, sa_expr)
            return
        elif isinstance(expr, Mod):
            lhs, rhs = self._expr_to_sqlalchemy[expr._lhs], self._expr_to_sqlalchemy[expr._rhs]
            sa_expr = BINARY_OP[expr.node_name](lhs, rhs)
            if not is_constant_scalar(expr._rhs):
                sa_expr = case([(rhs > 0, func.abs(sa_expr))], else_=sa_expr)
            elif expr._rhs.value > 0:
                sa_expr = func.abs(sa_expr)
            self._add(expr, sa_expr)
            return
        else:
            op = BINARY_OP[expr.node_name]
        lhs, rhs = self._expr_to_sqlalchemy[expr._lhs], self._expr_to_sqlalchemy[expr._rhs]
        sa_expr = op(lhs, rhs)
        self._add(expr, sa_expr)

    def visit_unary_op(self, expr):
        if isinstance(expr, Abs):
            op = func.abs
        else:
            op = UNARY_OP[expr.node_name]
        self._add(expr, op(self._expr_to_sqlalchemy[expr._input]))

    def visit_math(self, expr):
        try:
            op = getattr(func, MATH_COMPILE_DIC[expr.node_name])
            sa_expr = op(self._expr_to_sqlalchemy[expr._input])
        except KeyError:
            if expr.node_name == 'Log':
                if expr._base is not None:
                    sa_expr = SALog('log', self._expr_to_sqlalchemy[expr._base],
                                    self._expr_to_sqlalchemy[expr._base],
                                    self._expr_to_sqlalchemy[expr._input])
                else:
                    sa_expr = SALog('log', None, self._expr_to_sqlalchemy[expr._input])
            elif expr.node_name == 'Log2':
                sa_expr = SALog('log', 2, 2, self._expr_to_sqlalchemy[expr._input])
                sa_expr = sa_expr.cast(types.df_type_to_sqlalchemy_type(expr.dtype))
            elif expr.node_name == 'Log10':
                sa_expr = SALog('log', 10, 10, self._expr_to_sqlalchemy[expr._input])
                sa_expr = sa_expr.cast(types.df_type_to_sqlalchemy_type(expr.dtype))
            elif expr.node_name == 'Trunc':
                input = self._expr_to_sqlalchemy[expr._input]
                decimals = 0 if expr._decimals is None else self._expr_to_sqlalchemy[expr._decimals]
                sa_expr = SATruncate('trunc', input, decimals)
            elif expr.node_name == 'Round':
                decimals = 0 if expr._decimals is None else self._expr_to_sqlalchemy[expr._decimals]
                sa_expr = func.round(self._expr_to_sqlalchemy[expr._input], decimals)
            else:
                raise NotImplementedError

        self._add(expr, sa_expr)

    def visit_string_op(self, expr):
        if isinstance(expr, strings.Capitalize):
            input = self._expr_to_sqlalchemy[expr._input]
            tp = types.df_type_to_sqlalchemy_type(expr.dtype)
            sa_expr = func.upper(func.substr(input, 1, 1)).cast(tp) + \
                      func.lower(func.substr(input, 2)).cast(tp)
        elif isinstance(expr, strings.Contains) and not expr.regex:
            sa_expr = self._expr_to_sqlalchemy[expr._input].contains(
                self._expr_to_sqlalchemy[expr._pat])
        elif isinstance(expr, strings.Endswith):
            sa_expr = self._expr_to_sqlalchemy[expr._input].endswith(
                self._expr_to_sqlalchemy[expr._pat])
        elif isinstance(expr, strings.Startswith):
            sa_expr = self._expr_to_sqlalchemy[expr._input].startswith(
                self._expr_to_sqlalchemy[expr._pat])
        elif isinstance(expr, strings.Replace) and not expr.regex:
            sa_expr = func.replace(self._expr_to_sqlalchemy[expr._input],
                                   self._expr_to_sqlalchemy[expr._pat],
                                   self._expr_to_sqlalchemy[expr._repl])
        elif isinstance(expr, strings.Get):
            sa_expr = func.substr(self._expr_to_sqlalchemy[expr._input],
                                  self._expr_to_sqlalchemy[expr._index] + 1, 1)
        elif isinstance(expr, strings.Len):
            sa_expr = func.length(self._expr_to_sqlalchemy[expr._input])
        elif isinstance(expr, (strings.Ljust, strings.Rjust, strings.Pad)):
            if isinstance(expr, strings.Pad):
                if expr.side == 'both':
                    raise NotImplementedError
                op = func.lpad if expr.side == 'left' else func.rpad
            else:
                op = func.lpad if isinstance(expr, strings.Ljust) else func.rpad
            sa_expr = op(self._expr_to_sqlalchemy[expr._input],
                         self._expr_to_sqlalchemy[expr._width],
                         self._expr_to_sqlalchemy[expr._fillchar])
        elif isinstance(expr, (strings.Lower, strings.Upper)):
            op = func.lower if isinstance(expr, strings.Lower) else func.upper
            sa_expr = op(self._expr_to_sqlalchemy[expr._input])
        elif isinstance(expr, (strings.Lstrip, strings.Rstrip, strings.Strip)):
            if expr._to_strip is None:
                raise NotImplementedError
            op = func.ltrim if isinstance(expr, strings.Lstrip) else (
                func.rtrim if isinstance(expr, strings.Rstrip) else func.btrim
            )
            sa_expr = op(self._expr_to_sqlalchemy[expr._input],
                         self._expr_to_sqlalchemy[expr._to_strip])
        elif isinstance(expr, strings.Repeat):
            sa_expr = func.repeat(self._expr_to_sqlalchemy[expr._input],
                                  self._expr_to_sqlalchemy[expr._repeats])
        elif isinstance(expr, strings.Slice):
            if expr.end is None and expr.step is None:
                sa_expr = func.substr(self._expr_to_sqlalchemy[expr._input],
                                      self._expr_to_sqlalchemy[expr._start] + 1)
            elif isinstance(expr.start, six.integer_types) and \
                    isinstance(expr.end, six.integer_types) and \
                    expr.step is None and expr.start > 0 and expr.end > 0:
                length = expr.end - expr.start
                sa_expr = func.substr(self._expr_to_sqlalchemy[expr._input],
                                      expr.start + 1, length)
            else:
                raise NotImplementedError
        elif isinstance(expr, strings.Title):
            sa_expr = func.initcap(self._expr_to_sqlalchemy[expr._input])
        else:
            raise NotImplementedError

        self._add(expr, sa_expr)

    def visit_datetime_op(self, expr):
        class_name = type(expr).__name__
        input = self._expr_to_sqlalchemy[expr._input]

        if class_name in DATE_PARTS_DIC:
            if self._sa_engine and self._sa_engine.name == 'mysql':
                if class_name == 'UnixTimestamp':
                    fun = func.unix_timestamp
                else:
                    fun = getattr(func, class_name.lower())
                sa_expr = fun(input).cast(types.df_type_to_sqlalchemy_type(expr.dtype))
            else:
                sa_expr = func.date_part(DATE_PARTS_DIC[class_name], input)\
                    .cast(types.df_type_to_sqlalchemy_type(expr.dtype))
        elif isinstance(expr, Date):
            if self._sa_engine and self._sa_engine.name == 'mysql':
                sa_expr = func.date(input).cast(types.df_type_to_sqlalchemy_type(expr.dtype))
            else:
                sa_expr = func.date_trunc('day', input)
        elif isinstance(expr, WeekDay):
            if self._sa_engine and self._sa_engine.name == 'mysql':
                sa_expr = (func.dayofweek(input).cast(types.df_type_to_sqlalchemy_type(expr.dtype)) + 5) % 7
            else:
                sa_expr = (func.date_part('dow', input).cast(types.df_type_to_sqlalchemy_type(expr.dtype)) + 6) % 7
        else:
            raise NotImplementedError

        self._add(expr, sa_expr)

    def visit_groupby(self, expr):
        bys, having, aggs, fields = tuple(expr.args[1:])
        if fields is None:
            fields = bys + aggs

        selects = select(self._gen_select_columns(fields))
        if len(fields) == 1 and isinstance(fields[0], (Count, GroupedCount)):
            selects = selects.select_from(self._expr_to_sqlalchemy[fields[0].input])
        grouped = selects.group_by(*self._gen_select_columns(bys))
        if having:
            grouped = grouped.having(self._expr_to_sqlalchemy[having])

        if expr is not self._expr_dag.root:
            grouped = grouped.alias(self._new_alias())

        self._add(expr, grouped)

    def visit_mutate(self, expr):
        bys, mutates, fields = tuple(expr.args[1:])
        if fields is None:
            fields = bys + mutates

        selects = select(self._gen_select_columns(fields))
        if expr is not self._expr_dag.root:
            selects = selects.alias(self._new_alias())

        self._add(expr, selects)

    def visit_sort_column(self, expr):
        if isinstance(expr.input, CollectionExpr):
            sa_expr = self._expr_to_sqlalchemy[expr.input].c[expr.source_name]
        else:
            sa_expr = self._expr_to_sqlalchemy[expr.input]
        if not expr._ascending:
            sa_expr = desc(sa_expr)

        self._add(expr, sa_expr)

    def visit_sort(self, expr):
        input = self._expr_to_sqlalchemy[expr.input]
        sa_expr = input.select(order_by=[self._expr_to_sqlalchemy[e]
                                         for e in expr._sorted_fields])
        if expr is not self._expr_dag.root:
            sa_expr = sa_expr.alias(self._new_alias())
        self._add(expr, sa_expr)

    def visit_distinct(self, expr):
        sa_expr = select(self._gen_select_columns(expr._unique_fields), distinct=True)

        if expr is not self._expr_dag.root:
            sa_expr = sa_expr.alias(self._new_alias())
        self._add(expr, sa_expr)

    def visit_column(self, expr):
        table = self._expr_to_sqlalchemy[expr.input]
        col = table.c[expr.source_name]

        if expr._source_data_type != expr._data_type:
            col = col.cast(types.df_type_to_sqlalchemy_type(expr._data_type))

        self._add(expr, col)

    def visit_reduction(self, expr):
        if getattr(expr, '_unique', False):
            raise NotImplementedError

        input = self._expr_to_sqlalchemy[expr.input]

        # TODO: MEDIAN does not support
        if isinstance(expr, (Max, GroupedMax)):
            f = func.max
        elif isinstance(expr, (Min, GroupedMin)):
            f = func.min
        elif isinstance(expr, (Count, GroupedCount)):
            f = func.count
        elif isinstance(expr, (Sum, GroupedSum)):
            f = func.sum
        elif isinstance(expr, (Var, GroupedVar)) and expr._ddof in (0, 1):
            f = func.var_pop if expr._ddof == 0 else func.var_samp
        elif isinstance(expr, (Std, GroupedStd)) and expr._ddof in (0, 1):
            f = func.stddev_pop if expr._ddof == 0 else func.stddev_samp
        elif isinstance(expr, (Mean, GroupedMean)):
            f = func.avg
        elif isinstance(expr, (NUnique, GroupedNUnique)):
            f = lambda *x: func.count(distinct(*x))
        elif isinstance(expr, (Cat, GroupedCat)):
            f = lambda x: func.array_to_string(func.array_agg(x),
                                               self._expr_to_sqlalchemy[expr._sep])
        else:
            raise NotImplementedError

        if isinstance(expr, (Count, GroupedCount)) and \
                isinstance(expr.input, CollectionExpr):
            reduced = f()
        elif isinstance(expr, (NUnique, GroupedNUnique)):
            if len(expr.inputs) > 1:
                raise NotImplementedError
            reduced = f(*(self._expr_to_sqlalchemy[i] for i in expr.inputs))
        else:
            reduced = f(input)
        self._add(expr, reduced)

    def visit_cum_window(self, expr):
        input = self._expr_to_sqlalchemy[expr._input]
        if expr._distinct.value is True:
            raise NotImplementedError
        try:
            func_name = WINDOW_COMPILE_DIC[expr.node_name]
        except KeyError:
            raise NotImplementedError
        f = getattr(func, func_name)
        partition_by = self._gen_select_columns(expr._partition_by) \
            if expr._partition_by else None
        order_by = self._gen_select_columns(expr._order_by) \
            if expr._order_by else None
        rows = (self._expr_to_sqlalchemy[expr._preceding] if expr._preceding else None,
                self._expr_to_sqlalchemy[expr._following] if expr._following else None)
        rows = None if all(r is None for r in rows) else rows

        sa_expr = f(input).over(partition_by=partition_by, order_by=order_by, rows=rows)
        self._add(expr, sa_expr)

    def visit_rank_window(self, expr):
        try:
            func_name = WINDOW_COMPILE_DIC[expr.node_name]
        except KeyError:
            raise NotImplementedError
        f = getattr(func, func_name)
        partition_by = self._gen_select_columns(expr._partition_by) \
            if expr._partition_by else None
        order_by = self._gen_select_columns(expr._order_by) \
            if expr._order_by else None

        sa_expr = f().over(partition_by=partition_by, order_by=order_by)
        if isinstance(expr, PercentRank):
            sa_expr = sa_expr.cast(types.df_type_to_sqlalchemy_type(expr.dtype))
        self._add(expr, sa_expr)

    def visit_shift_window(self, expr):
        input = self._expr_to_sqlalchemy[expr._input]
        try:
            func_name = WINDOW_COMPILE_DIC[expr.node_name]
        except KeyError:
            raise NotImplementedError
        f = getattr(func, func_name)
        partition_by = self._gen_select_columns(expr._partition_by) \
            if expr._partition_by else None
        order_by = self._gen_select_columns(expr._order_by) \
            if expr._order_by else None

        args = (input, self._expr_to_sqlalchemy[expr._offset])
        if expr._default:
            args += (literal(self._expr_to_sqlalchemy[expr._default]).cast(
                types.df_type_to_sqlalchemy_type(expr._input.dtype)),)

        sa_expr = f(*args).over(partition_by=partition_by, order_by=order_by)
        self._add(expr, sa_expr)

    def visit_scalar(self, expr):
        if expr._value is not None:
            if expr.dtype == df_types.string:
                val = utils.to_str(expr.value) \
                    if isinstance(expr.value, six.text_type) else expr.value
                self._add(expr, val)
                return
            else:
                self._add(expr, expr._value)
        else:
            self._add(expr, None)

    def visit_cast(self, expr):
        to_type = types.df_type_to_sqlalchemy_type(expr.dtype)
        self._add(expr, self._expr_to_sqlalchemy[expr.input].cast(to_type))

    def visit_join(self, expr):
        lhs, rhs = self._expr_to_sqlalchemy[expr._lhs], self._expr_to_sqlalchemy[expr._rhs]
        if isinstance(expr, RightJoin):
            lhs, rhs = rhs, lhs
        on = self._expr_to_sqlalchemy[expr._predicate]
        kw = dict()
        if isinstance(expr, OuterJoin):
            kw['full'] = True
        elif isinstance(expr, (LeftJoin, RightJoin)):
            kw['isouter'] = True
        joined = join(lhs, rhs, onclause=on, **kw)

        self._add(expr, joined)

    def visit_union(self, expr):
        lhs, rhs = self._expr_to_sqlalchemy[expr._lhs], self._expr_to_sqlalchemy[expr._rhs]
        if is_source_collection(expr._lhs):
            lhs = select([lhs])
        elif isinstance(lhs, Alias):
            lhs = lhs.element
        if is_source_collection(expr._rhs):
            rhs = select([rhs])
        elif isinstance(rhs, Alias):
            rhs = rhs.element
        method = union if expr._distinct else union_all
        unioned = method(lhs, rhs)

        if expr is not self._expr_dag.root:
            unioned = unioned.alias(self._new_alias())

        self._add(expr, unioned)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..compat import six
from .. import serializers

LOG_TYPES_MAPPING = {
    'hs_err_log': 'hs_err_*.log',
    'coreinfo': 'coreinfo.tmp',
}
LOG_TYPES_MAPPING.update({k: k for k in 'stdout stderr waterfall_summary jstack pstack'.split()})


class Worker(serializers.JSONSerializableModel):
    """
    Worker information class for worker information and log retrieval.
    """
    __slots__ = '_client',

    @classmethod
    def extract_from_json(cls, json_obj, client=None, parent=None):
        raise NotImplementedError

    def get_log(self, log_type, size=0):
        """
        Get logs from worker.

        :param log_type: type of logs. Possible log types contains {log_types}
        :param size: length of the log to retrieve
        :return: log content
        """
        return self.parent.get_worker_log(self.log_id, log_type, size=size)
    get_log.__doc__ = get_log.__doc__.format(log_types=', '.join(sorted(six.iterkeys(LOG_TYPES_MAPPING))))


class WorkerDetail2(Worker):
    id = serializers.JSONNodeField('id')
    log_id = serializers.JSONNodeField('logId')
    type = serializers.JSONNodeField('type')
    start_time = serializers.JSONNodeField('startTime', parse_callback=int)
    end_time = serializers.JSONNodeField('endTime', parse_callback=int)
    status = serializers.JSONNodeField('status')
    gbi_counter = serializers.JSONNodeField('gblCounter')
    input_bytes = serializers.JSONNodeField('input_bytes', parse_callback=int)
    input_records = serializers.JSONNodeField('input_records', parse_callback=int)
    output_bytes = serializers.JSONNodeField('output_bytes', parse_callback=int)
    output_records = serializers.JSONNodeField('output_records', parse_callback=int)

    @classmethod
    def extract_from_json(cls, json_obj, client=None, parent=None):
        workers = []

        def _extract(o):
            if isinstance(o, list):
                for v in o:
                    _extract(v)
            elif isinstance(o, dict):
                worker_type = o.get('name', '')
                if 'instances' in o:
                    for v in o['instances']:
                        w = cls.parse(v, parent=parent)
                        w.type = worker_type
                        w._client = client
                        workers.append(w)
                    return
                for _, v in six.iteritems(o):
                    _extract(v)

        _extract(json_obj)
        return workers



# Copyright 1999-2024 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import serializers
from ..compat import Enum


class ClusterType(Enum):
    HASH = "hash"
    RANGE = "range"


class ClusterSortOrder(Enum):
    ASC = "ASC"
    DESC = "DESC"


class ClusterSortCol(serializers.JSONSerializableModel):
    name = serializers.JSONNodeField('col')
    order = serializers.JSONNodeField(
        'order', parse_callback=lambda x: ClusterSortOrder(x.upper()) if x else None
    )


class ClusterInfo(serializers.JSONSerializableModel):
    cluster_type = serializers.JSONNodeField(
        'ClusterType', parse_callback=lambda x: ClusterType(x.lower()) if x else None
    )
    bucket_num = serializers.JSONNodeField('BucketNum')
    cluster_cols = serializers.JSONNodeField('ClusterCols')
    sort_cols = serializers.JSONNodesReferencesField(ClusterSortCol, 'SortCols')

    @classmethod
    def deserial(cls, content, obj=None, **kw):
        res = super(ClusterInfo, cls).deserial(content, obj=obj, **kw)
        return res if res.cluster_type is not None else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import copy
import itertools
import math
from collections import OrderedDict

from requests import Response

from . import types, compat, utils, options
from .models.record import Record
from .compat import six, StringIO


class AbstractRecordReader(object):

    def __iter__(self):
        return self

    def __next__(self):
        raise NotImplementedError

    next = __next__

    @classmethod
    def _calc_count(cls, start, end, step):
        if end is None:
            return end
        step = step or 1
        return int(math.ceil(float(end - start) / step))

    @classmethod
    def _get_slice(cls, item):
        if isinstance(item, six.integer_types):
            start = item
            end = start + 1
            step = 1
        elif isinstance(item, slice):
            start = item.start or 0
            end = item.stop
            step = item.step or 1
        else:
            raise ValueError('Reader only supports index and slice operation.')

        return start, end, step

    def __getitem__(self, item):
        start, end, step = self._get_slice(item)
        count = self._calc_count(start, end, step)

        if start < 0 or (count is not None and count <= 0) or step < 0:
            raise ValueError('start, count, or step cannot be negative')

        it = self._get_slice_iter(start=start, end=end, step=step)
        if isinstance(item, six.integer_types):
            try:
                return next(it)
            except StopIteration:
                raise IndexError('Index out of range: %s' % item)
        return it

    def _get_slice_iter(self, start=None, end=None, step=None):
        class SliceIterator(six.Iterator):
            def __init__(self, it):
                self.it = it

            def __iter__(self):
                return self.it

            def __next__(self):
                return next(self.it)

            @staticmethod
            def to_pandas():
                if end is not None:
                    count = (end - (start or 0)) // (step or 1)
                else:
                    count = None
                pstep = None if step == 1 else step
                kw = dict(start=start, count=count, step=pstep)
                kw = {k: v for k, v in kw.items() if v is not None}
                return parent.to_pandas(**kw)

        parent = self
        return SliceIterator(self._iter(start=start, end=end, step=step))

    def _iter(self, start=None, end=None, step=None):
        start = start or 0
        step = step or 1
        curr = start

        for _ in range(start):
            try:
                next(self)
            except StopIteration:
                return

        while True:
            for i in range(step):
                try:
                    record = next(self)
                except StopIteration:
                    return
                if i == 0:
                    yield record
                curr += 1
                if end is not None and curr >= end:
                    return

    def _data_to_result_frame(self, data, unknown_as_string=True, as_type=None):
        from .df.backends.frame import ResultFrame
        from .df.backends.odpssql.types import odps_schema_to_df_schema, odps_type_to_df_type

        kw = dict()
        if getattr(self, 'schema', None) is not None:
            kw['schema'] = odps_schema_to_df_schema(self.schema)
        elif getattr(self, '_schema', None) is not None:
            # do not remove as there might be coverage missing
            kw['schema'] = odps_schema_to_df_schema(self._schema)

        if getattr(self, '_column_names', None) is not None:
            self._columns = [self.schema[c] for c in self._column_names]
        if getattr(self, '_columns', None) is not None:
            cols = []
            for col in self._columns:
                col = copy.copy(col)
                col.type = odps_type_to_df_type(col.type)
                cols.append(col)
            kw['columns'] = cols

        if hasattr(self, 'raw'):
            try:
                import pandas as pd
                from .df.backends.pd.types import pd_to_df_schema
                data = pd.read_csv(StringIO(self.raw))
                kw['schema'] = pd_to_df_schema(
                    data, unknown_as_string=unknown_as_string, as_type=as_type
                )
                kw.pop('columns', None)
            except ImportError:
                pass

        if not kw:
            raise ValueError('Cannot convert to ResultFrame from %s.' % type(self).__name__)

        return ResultFrame(data, **kw)

    def to_result_frame(
        self, unknown_as_string=True, as_type=None, start=None, count=None, **iter_kw
    ):
        read_row_batch_size = options.tunnel.read_row_batch_size
        if "end" in iter_kw:
            end = iter_kw["end"]
        else:
            end = None if count is None else (start or 0) + count * (iter_kw.get("step") or 1)

        frames = []
        if hasattr(self, "raw"):
            # data represented as raw csv: just skip iteration
            data = [r for r in self._iter(start=start, end=end, **iter_kw)]
        else:
            offset_iter = itertools.cycle(compat.irange(read_row_batch_size))
            data = [None] * read_row_batch_size
            for offset, rec in zip(
                offset_iter, self._iter(start=start, end=end, **iter_kw)
            ):
                data[offset] = rec
                if offset != read_row_batch_size - 1:
                    continue

                frames.append(self._data_to_result_frame(
                    data, unknown_as_string=unknown_as_string, as_type=as_type
                ))
                data = [None] * read_row_batch_size
                if len(frames) > options.tunnel.batch_merge_threshold:
                    frames = [frames[0].concat(*frames[1:])]

        if not frames or data[0] is not None:
            data = list(itertools.takewhile(lambda x: x is not None, data))
            frames.append(self._data_to_result_frame(
                data, unknown_as_string=unknown_as_string, as_type=as_type
            ))
        return frames[0].concat(*frames[1:])

    def to_pandas(self, start=None, count=None, **kw):
        import pandas  # noqa: F401

        return self.to_result_frame(start=start, count=count, **kw).values


class CsvRecordReader(AbstractRecordReader):
    NULL_TOKEN = '\\N'
    BACK_SLASH_ESCAPE = '\\x%02x' % ord('\\')

    def __init__(self, schema, stream, **kwargs):
        self._schema = schema
        self._columns = None
        self._fp = stream
        if isinstance(self._fp, Response):
            self.raw = self._fp.content if six.PY2 else self._fp.text
        else:
            self.raw = self._fp

        if options.tunnel.string_as_binary:
            self._csv = csv.reader(six.StringIO(self._escape_csv_bin(self.raw)))
        else:
            self._csv = csv.reader(six.StringIO(self._escape_csv(self.raw)))

    @classmethod
    def _escape_csv(cls, s):
        escaped = utils.to_text(s).encode('unicode_escape')
        # Make invisible chars available to `csv` library.
        # Note that '\n' and '\r' should be unescaped.
        # '\\' should be replaced with '\x5c' before unescaping
        # to avoid mis-escaped strings like '\\n'.
        return utils.to_text(escaped) \
            .replace('\\\\', cls.BACK_SLASH_ESCAPE) \
            .replace('\\n', '\n') \
            .replace('\\r', '\r')

    @classmethod
    def _escape_csv_bin(cls, s):
        escaped = utils.to_binary(s).decode('latin1').encode('unicode_escape')
        # Make invisible chars available to `csv` library.
        # Note that '\n' and '\r' should be unescaped.
        # '\\' should be replaced with '\x5c' before unescaping
        # to avoid mis-escaped strings like '\\n'.
        return utils.to_text(escaped) \
            .replace('\\\\', cls.BACK_SLASH_ESCAPE) \
            .replace('\\n', '\n') \
            .replace('\\r', '\r')

    @staticmethod
    def _unescape_csv(s):
        return s.encode('utf-8').decode('unicode_escape')

    @staticmethod
    def _unescape_csv_bin(s):
        return s.encode('utf-8').decode('unicode_escape').encode('latin1')

    def _readline(self):
        try:
            values = next(self._csv)
            res = []

            read_binary = options.tunnel.string_as_binary
            if read_binary:
                unescape_csv = self._unescape_csv_bin
            else:
                unescape_csv = self._unescape_csv

            for i, value in enumerate(values):
                value = unescape_csv(value)
                if value == self.NULL_TOKEN:
                    res.append(None)
                elif self._columns and self._columns[i].type == types.boolean:
                    if value == 'true':
                        res.append(True)
                    elif value == 'false':
                        res.append(False)
                    else:
                        res.append(value)
                elif self._columns and isinstance(self._columns[i].type, types.Map):
                    col_type = self._columns[i].type
                    if not (value.startswith('{') and value.endswith('}')):
                        raise ValueError('Dict format error!')

                    items = []
                    for kv in value[1:-1].split(','):
                        k, v = kv.split(':', 1)
                        k = col_type.key_type.cast_value(k.strip(), types.string)
                        v = col_type.value_type.cast_value(v.strip(), types.string)
                        items.append((k, v))
                    res.append(OrderedDict(items))
                elif self._columns and isinstance(self._columns[i].type, types.Array):
                    col_type = self._columns[i].type
                    if not (value.startswith('[') and value.endswith(']')):
                        raise ValueError('Array format error!')

                    items = []
                    for item in value[1:-1].split(','):
                        item = col_type.value_type.cast_value(item.strip(), types.string)
                        items.append(item)
                    res.append(items)
                else:
                    res.append(value)
            return res
        except StopIteration:
            return

    def __next__(self):
        self._load_columns()

        values = self._readline()
        if values is None:
            raise StopIteration
        return Record(self._columns, values=values)

    next = __next__

    def read(self, start=None, count=None, step=None):
        if count is None:
            end = None
        else:
            start = start or 0
            step = step or 1
            end = start + count * step
        return self._iter(start=start, end=end, step=step)

    def _load_columns(self):
        if self._columns is not None:
            return

        values = self._readline()
        self._columns = []
        for value in values:
            if self._schema is None:
                self._columns.append(types.Column(name=value, typo='string'))
            else:
                if self._schema.is_partition(value):
                    self._columns.append(self._schema.get_partition(value))
                else:
                    self._columns.append(self._schema.get_column(value))

    def close(self):
        if hasattr(self._fp, 'close'):
            self._fp.close()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


# make class name compatible
RecordReader = CsvRecordReader



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cupid/proto/cupid_process_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cupid/proto/cupid_process_service.proto',
  package='apsara.odps.cupid.protocol',
  syntax='proto2',
  serialized_options=_b('B\030CupidProcessServiceProto\200\001\001\210\001\001\220\001\001'),
  serialized_pb=_b('\n\'cupid/proto/cupid_process_service.proto\x12\x1a\x61psara.odps.cupid.protocol\"\'\n\x08\x45nvEntry\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"A\n\x08\x43hildEnv\x12\x35\n\x07\x65ntries\x18\x01 \x03(\x0b\x32$.apsara.odps.cupid.protocol.EnvEntry2g\n\x0eProcessService\x12U\n\x07Prepare\x12$.apsara.odps.cupid.protocol.ChildEnv\x1a$.apsara.odps.cupid.protocol.ChildEnvB#B\x18\x43upidProcessServiceProto\x80\x01\x01\x88\x01\x01\x90\x01\x01')
)




_ENVENTRY = _descriptor.Descriptor(
  name='EnvEntry',
  full_name='apsara.odps.cupid.protocol.EnvEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='apsara.odps.cupid.protocol.EnvEntry.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='apsara.odps.cupid.protocol.EnvEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=71,
  serialized_end=110,
)


_CHILDENV = _descriptor.Descriptor(
  name='ChildEnv',
  full_name='apsara.odps.cupid.protocol.ChildEnv',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entries', full_name='apsara.odps.cupid.protocol.ChildEnv.entries', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=112,
  serialized_end=177,
)

_CHILDENV.fields_by_name['entries'].message_type = _ENVENTRY
DESCRIPTOR.message_types_by_name['EnvEntry'] = _ENVENTRY
DESCRIPTOR.message_types_by_name['ChildEnv'] = _CHILDENV
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EnvEntry = _reflection.GeneratedProtocolMessageType('EnvEntry', (_message.Message,), dict(
  DESCRIPTOR = _ENVENTRY,
  __module__ = 'cupid.proto.cupid_process_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.EnvEntry)
  ))
_sym_db.RegisterMessage(EnvEntry)

ChildEnv = _reflection.GeneratedProtocolMessageType('ChildEnv', (_message.Message,), dict(
  DESCRIPTOR = _CHILDENV,
  __module__ = 'cupid.proto.cupid_process_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.ChildEnv)
  ))
_sym_db.RegisterMessage(ChildEnv)


DESCRIPTOR._options = None

_PROCESSSERVICE = _descriptor.ServiceDescriptor(
  name='ProcessService',
  full_name='apsara.odps.cupid.protocol.ProcessService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=179,
  serialized_end=282,
  methods=[
  _descriptor.MethodDescriptor(
    name='Prepare',
    full_name='apsara.odps.cupid.protocol.ProcessService.Prepare',
    index=0,
    containing_service=None,
    input_type=_CHILDENV,
    output_type=_CHILDENV,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_PROCESSSERVICE)

DESCRIPTOR.services_by_name['ProcessService'] = _PROCESSSERVICE

ProcessService = service_reflection.GeneratedServiceType('ProcessService', (_service.Service,), dict(
  DESCRIPTOR = _PROCESSSERVICE,
  __module__ = 'cupid.proto.cupid_process_service_pb2'
  ))

ProcessService_Stub = service_reflection.GeneratedServiceStubType('ProcessService_Stub', (ProcessService,), dict(
  DESCRIPTOR = _PROCESSSERVICE,
  __module__ = 'cupid.proto.cupid_process_service_pb2'
  ))


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple, OrderedDict

try:
    string_types = (unicode, basestring)
    iteritems = lambda d: d.iteritems()
except:
    string_types = (bytes, str)
    iteritems = lambda d: d.items()


class NamedTupleMixin(object):
    def __getattr__(self, item):
        if item in self._name_map:
            return self[self._name_map[item]]
        elif item == "get":
            return self._NamedTupleMixin_get
        elif item == "items":
            return self._NamedTupleMixin_items
        elif item == "keys":
            return self._NamedTupleMixin_keys
        elif item == "values":
            return self._NamedTupleMixin_values
        else:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, str(item))
            )

    def __getitem__(self, item):
        try:
            return self._base.__getitem__(self, item)
        except TypeError:
            pass

        return self[self._name_map[item]]

    def __eq__(self, other):
        if isinstance(other, tuple):
            return tuple(self) == other
        else:
            return self.asdict() == other

    def _NamedTupleMixin_get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def _NamedTupleMixin_items(self):
        for k in self._names:
            yield k, getattr(self, k)

    def _NamedTupleMixin_keys(self):
        return self._names

    def _NamedTupleMixin_values(self):
        for k in self._names:
            yield getattr(self, k)

    def asdict(self):
        return OrderedDict(zip(self._names, self))

    def replace(self, **kwds):
        new_kw = {self._fields[self._name_map[k]]: v for k, v in iteritems(kwds)}
        return self._replace(**new_kw)


def xnamedtuple(typename, field_names):
    if isinstance(field_names, string_types):
        field_names = field_names.replace(",", " ").split()
    base_nt = namedtuple(typename + "_base", field_names, rename=True)
    nt = type(typename, (NamedTupleMixin, base_nt), {})
    nt._base = base_nt
    nt._name_map = {v: k for k, v in enumerate(field_names)}
    nt._names = field_names
    return nt



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

include "util_c.pxi"

from libc.stdint cimport *
from libc.string cimport *
from libcpp.string cimport string
from ...src.stringstream cimport stringstream


cdef class CEncoder:
    def __cinit__(self, size_t reserve):
        cdef string* prealloc_string = new string()
        prealloc_string.reserve(reserve)
        self._buffer = new stringstream(prealloc_string[0])
        del prealloc_string

    def __dealloc__(self):
        del self._buffer

    cdef size_t position(self) nogil:
        return self._buffer.tellp()

    cpdef bytes tostring(self):
        return bytes(self._buffer.to_string())

    cdef int append_tag(self, int field_num, int wire_type) except + nogil:
        cdef int key
        key = (field_num << 3) | wire_type
        cdef int size = set_varint64(key, self._buffer[0])
        return size

    cdef int append_sint32(self, int32_t value) except + nogil:
        return set_signed_varint32(value, self._buffer[0])

    cdef int append_uint32(self, uint32_t value) except + nogil:
        return set_varint32(value, self._buffer[0])

    cdef int append_sint64(self, int64_t value) except + nogil:
        return set_signed_varint64(value, self._buffer[0])

    cdef int append_uint64(self, uint64_t value) except + nogil:
        return set_varint64(value, self._buffer[0])

    cdef int append_bool(self, bint value) except + nogil:
        return set_varint32(value, self._buffer[0])

    cdef int append_float(self, float value) except + nogil:
        self._buffer.write(<const char *>&value, sizeof(float))
        return sizeof(float)

    cdef int append_double(self, double value) except + nogil:
        self._buffer.write(<const char*>&value, sizeof(double))
        return sizeof(double)

    cdef int append_string(self, const char *ptr, size_t value_len) except + nogil:
        cdef int size = set_varint32(value_len, self._buffer[0])
        self._buffer.write(ptr, value_len)
        return size + value_len


cdef class Encoder:
    def __init__(self):
        self._encoder = CEncoder(4096)

    def __len__(self):
        return self._encoder.position()

    def position(self):
        return self._encoder.position()

    def tostring(self):
        return self._encoder.tostring()

    def append_tag(self, int field_num, int wire_type):
        return self._encoder.append_tag(field_num, wire_type)

    def append_sint32(self, int32_t value):
        return self._encoder.append_sint32(value)

    def append_uint32(self, uint32_t value):
        return self._encoder.append_uint32(value)

    def append_sint64(self, int64_t value):
        return self._encoder.append_sint64(value)

    def append_uint64(self, uint64_t value):
        return self._encoder.append_uint64(value)

    def append_bool(self, bint value):
        return self._encoder.append_bool(value)

    def append_float(self, float value):
        self._encoder.append_float(value)

    def append_double(self, double value):
        self._encoder.append_double(value)

    def append_string(self, bytes value):
        return self._encoder.append_string(value, len(value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
import math as pymath

from ..analyzer import BaseAnalyzer
from ...expr.arithmetic import *
from ...expr.composites import ListDictGetItem
from ...expr.math import *
from ...expr.datetimes import *
from ...expr.strings import *
from ...expr.strings import Count as StrCount
from ...expr.element import *
from ...expr.reduction import *
from ...expr.collections import *
from ...expr.merge import *
from ...expr.window import QCut
from ...utils import output
from ..errors import CompileError
from ..utils import refresh_dynamic
from ... import types
from .... import compat
from ....utils import to_text

_NAN = float('nan')


class Analyzer(BaseAnalyzer):
    def _parents(self, expr):
        return self._dag.successors(expr)

    def visit_composite_op(self, expr):
        if isinstance(expr, ListDictGetItem) and isinstance(expr.input.dtype, types.List):
            if is_constant_scalar(expr._negative_handled) and expr._negative_handled.value:
                return

            key_expr = expr._key
            if is_constant_scalar(key_expr):
                if key_expr.value >= 0:
                    return
                sub = expr.input[expr.input.len() - (-key_expr.value)]
                sub._negative_handled = Scalar(True)
            else:
                expr._negative_handled = Scalar(True)
                neg_expr = expr.input[expr.input.len() + key_expr]
                neg_expr._negative_handled = Scalar(True)
                sub = (key_expr >= 0).ifelse(expr, neg_expr).rename(expr.name)
            self._sub(expr, sub)
        else:
            raise NotImplementedError

    @staticmethod
    def _make_isna_expr(expr, invert=False):
        from ... import func

        input_expr = expr.input
        has_float_col = (
            isinstance(input_expr, CollectionExpr)
            and any(isinstance(col.type, types.Float) for col in input_expr.columns)
        ) or (
            isinstance(input_expr, (SequenceExpr, Scalar))
            and isinstance(input_expr.dtype, types.Float)
        )

        if options.df.odps.nan_handler is None or not has_float_col:
            return input_expr.isnull() if not invert else input_expr.notnull()

        handler = options.df.odps.nan_handler.lower()
        if handler == "py":
            func = lambda x: pymath.isnan(x)
            func._identifier = "isnan"
            handle_fun = lambda x: x.map(func, rtype=bool)
        else:
            handle_fun = functools.partial(func.ISNAN, rtype=bool)

        if not invert:
            return input_expr.isnull() | handle_fun(input_expr)
        else:
            return input_expr.notnull() & ~handle_fun(input_expr)

    def visit_element_op(self, expr):
        if isinstance(expr, Between):
            if expr.inclusive:
                sub = ((expr.left <= expr.input) & (expr.input.copy() <= expr.right))
            else:
                sub = ((expr.left < expr.input) & (expr.input.copy() < expr.right))
            self._sub(expr, sub.rename(expr.name))
        elif isinstance(expr, Cut):
            sub = self._get_cut_sub_expr(expr)
            self._sub(expr, sub)
        elif isinstance(expr, IsNa):
            sub = self._make_isna_expr(expr)
            self._sub(expr, sub)
        elif isinstance(expr, NotNa):
            sub = self._make_isna_expr(expr, invert=True)
            self._sub(expr, sub)
        elif isinstance(expr, FillNa):
            cond = self._make_isna_expr(expr)
            sub = cond.ifelse(expr._fill_value, expr.input)
            if isinstance(expr, (SequenceExpr, Scalar)):
                sub = sub.rename(expr.name)
            self._sub(expr, sub)
        else:
            raise NotImplementedError

    def visit_sample(self, expr):
        if expr._replace.value or expr._weights is not None or expr._strata is not None:
            raise CompileError('ODPS SQL does not support specified sample method')
        if not expr._parts:
            raise NotImplementedError

        idxes = [None, ] if expr._i is None else expr._i

        condition = None
        for idx in idxes:
            inputs = [expr._parts]
            if idx is not None:
                new_val = idx.value + 1
                inputs.append(Scalar(_value=new_val, _value_type=idx.value_type))
            if expr._sampled_fields:
                inputs.extend(expr._sampled_fields)
            cond = MappedExpr(_inputs=inputs, _func='SAMPLE', _data_type=types.boolean)
            if condition is None:
                condition = cond
            else:
                condition |= cond
        sub = FilterCollectionExpr(_input=expr.input, _predicate=condition,
                                   _schema=expr.schema)
        expr.input.optimize_banned = True

        self._sub(expr, sub)

    def _visit_pivot(self, expr):
        sub = self._get_pivot_sub_expr(expr)
        self._sub(expr, sub)

    def _visit_pivot_table(self, expr):
        sub = self._get_pivot_table_sub_expr(expr)
        self._sub(expr, sub)

    def visit_pivot(self, expr):
        if isinstance(expr, PivotCollectionExpr):
            self._visit_pivot(expr)
        else:
            self._visit_pivot_table(expr)

    def visit_extract_kv(self, expr):
        kv_delimiter = expr._kv_delimiter.value
        item_delimiter = expr._item_delimiter.value
        default = expr._default.value if expr._default else None

        class KeyAgg(object):
            def buffer(self):
                return set()

            def __call__(self, buf, val):
                if not val:
                    return

                def validate_kv(v):
                    parts = v.split(kv_delimiter)
                    if len(parts) != 2:
                        raise ValueError('Malformed KV pair: %s' % v)
                    return parts[0]

                buf.update([validate_kv(item) for item in val.split(item_delimiter)])

            def merge(self, buf, pbuffer):
                buf.update(pbuffer)

            def getvalue(self, buf):
                return item_delimiter.join(sorted(buf))

        columns_expr = expr.input.exclude(expr._intact).apply(KeyAgg, names=[c.name for c in expr._columns])

        intact_names = [g.name for g in expr._intact]
        intact_types = [g.dtype for g in expr._intact]
        exprs = [expr]

        def callback(result, new_expr):
            expr = exprs[0]

            names = list(intact_names)
            tps = list(intact_types)
            kv_slot_map = dict()

            for col, key_str in compat.izip(result.columns, result[0]):
                kv_slot_map[col.name] = dict()
                for k in key_str.split(item_delimiter):
                    names.append('%s_%s' % (col.name, k))
                    tps.append(expr._column_type)
                    kv_slot_map[col.name][k] = len(names) - 1
            kv_slot_names = list(kv_slot_map.keys())

            type_adapter = None
            if isinstance(expr._column_type, types.Float):
                type_adapter = float
            elif isinstance(expr._column_type, types.Integer):
                type_adapter = int

            @output(names, tps)
            def mapper(row):
                ret = [default, ] * len(names)
                ret[:len(intact_names)] = [getattr(row, col) for col in intact_names]
                for col in kv_slot_names:
                    kv_val = getattr(row, col)
                    if not kv_val:
                        continue
                    for kv_item in kv_val.split(item_delimiter):
                        k, v = kv_item.split(kv_delimiter)
                        if type_adapter:
                            v = type_adapter(v)
                        ret[kv_slot_map[col][k]] = v
                return tuple(ret)

            new_expr._schema = TableSchema.from_lists(names, tps)

            extracted = expr.input.map_reduce(mapper)
            self._sub(new_expr, extracted)

            # trigger refresh of dynamic operations
            refresh_dynamic(extracted, self._dag)

        sub = CollectionExpr(_schema=DynamicSchema.from_lists(intact_names, intact_types),
                             _deps=[(columns_expr, callback)])
        self._sub(expr, sub)

    def visit_value_counts(self, expr):
        self._sub(expr, self._get_value_counts_sub_expr(expr))

    def _gen_mapped_expr(self, expr, inputs, func, name,
                         args=None, kwargs=None, multiple=False):
        kwargs = dict(_inputs=inputs, _func=func, _name=name,
                      _func_args=args, _func_kwargs=kwargs,
                      _multiple=multiple)
        if isinstance(expr, SequenceExpr):
            kwargs['_data_type'] = expr.dtype
        else:
            kwargs['_value_type'] = expr.dtype
        return MappedExpr(**kwargs)

    def visit_binary_op(self, expr):
        if not options.df.analyze:
            raise NotImplementedError

        if isinstance(expr, FloorDivide):
            func = lambda l, r: l // r
            # multiple False will pass *args instead of namedtuple
            sub = self._gen_mapped_expr(expr, (expr.lhs, expr.rhs),
                                        func, expr.name, multiple=False)
            self._sub(expr, sub)
            return
        if isinstance(expr, Mod):
            func = lambda l, r: l % r
            sub = self._gen_mapped_expr(expr, (expr.lhs, expr.rhs),
                                        func, expr.name, multiple=False)
            self._sub(expr, sub)
            return
        if isinstance(expr, Add) and \
                all(child.dtype == types.datetime for child in (expr.lhs, expr.rhs)):
            return
        elif isinstance(expr, (Add, Substract)):
            if expr.lhs.dtype == types.datetime and expr.rhs.dtype == types.datetime:
                pass
            elif any(isinstance(child, MilliSecondScalar) for child in (expr.lhs, expr.rhs)):
                pass
            else:
                return

            def func(l, r, method):
                from datetime import datetime, timedelta
                if not isinstance(l, datetime):
                    l = timedelta(milliseconds=l)
                if not isinstance(r, datetime):
                    r = timedelta(milliseconds=r)

                if method == '+':
                    res = l + r
                else:
                    res = l - r
                if isinstance(res, timedelta):
                    return int(res.total_seconds() * 1000)
                return res

            func._identifier = "dt_add_sub"
            inputs = expr.lhs, expr.rhs, Scalar('+') if isinstance(expr, Add) else Scalar('-')
            sub = self._gen_mapped_expr(expr, inputs, func, expr.name, multiple=False)
            self._sub(expr, sub)

        raise NotImplementedError

    def visit_unary_op(self, expr):
        if not options.df.analyze:
            raise NotImplementedError

        if isinstance(expr, Invert) and isinstance(expr.input.dtype, types.Integer):
            sub = expr.input.map(lambda x: ~x)
            self._sub(expr, sub)
            return

        raise NotImplementedError

    def visit_math(self, expr):
        if not options.df.analyze:
            raise NotImplementedError

        if expr.dtype != types.decimal:
            if isinstance(expr, Arccosh):
                def func(x):
                    return pymath.acosh(x)
            elif isinstance(expr, Arcsinh):
                def func(x):
                    return pymath.asinh(x)
            elif isinstance(expr, Arctanh):
                def func(x):
                    try:
                        return pymath.atanh(x)
                    except ValueError:
                        return _NAN
            elif isinstance(expr, Radians):
                def func(x):
                    return pymath.radians(x)
            elif isinstance(expr, Degrees):
                def func(x):
                    return pymath.degrees(x)
            else:
                raise NotImplementedError

            func._identifier = "m_" + type(expr).__name__.lower()
            sub = expr.input.map(func, expr.dtype)
            self._sub(expr, sub)
            return

        raise NotImplementedError

    def visit_datetime_op(self, expr):
        if isinstance(expr, Strftime):
            if not options.df.analyze:
                raise NotImplementedError

            def func(x, fmt):
                return x.strftime(fmt)

            func._identifier = "strftime"
            date_fmt = expr._date_format \
                if not isinstance(expr._date_format, StringScalar) or expr._date_format._value is None \
                else Scalar(to_text(expr.date_format).replace("%", "%%"))
            inputs = expr.input, date_fmt
            sub = self._gen_mapped_expr(expr, inputs, func, expr.name, multiple=False)
            self._sub(expr, sub)
            return

        raise NotImplementedError

    def visit_string_op(self, expr):
        if isinstance(expr, Ljust):
            rest = expr.width - expr.input.len()
            sub = expr.input + (rest >= 0).ifelse(expr._fillchar.repeat(rest), '')
            self._sub(expr, sub.rename(expr.name))
            return
        elif isinstance(expr, Rjust):
            rest = expr.width - expr.input.len()
            sub = (rest >= 0).ifelse(expr._fillchar.repeat(rest), '') + expr.input
            self._sub(expr, sub.rename(expr.name))
            return
        elif isinstance(expr, Zfill):
            fillchar = Scalar('0')
            rest = expr.width - expr.input.len()
            sub = (rest >= 0).ifelse(fillchar.repeat(rest), '') + expr.input
            self._sub(expr, sub.rename(expr.name))
            return
        elif isinstance(expr, CatStr):
            input = expr.input
            others = expr._others if isinstance(expr._others, Iterable) else (expr._others, )
            for other in others:
                if expr.na_rep is not None:
                    for e in (input, ) + tuple(others):
                        self._sub(e, e.fillna(expr.na_rep), parents=(expr, ))
                    return
                else:
                    if expr._sep is not None:
                        input = other.isnull().ifelse(input, input + expr._sep + other)
                    else:
                        input = other.isnull().ifelse(input, input + other)
            self._sub(expr, input.rename(expr.name))
            return

        if not options.df.analyze:
            raise NotImplementedError

        func = None
        if isinstance(expr, Contains) and expr.regex:
            def func(x, pat, case, flags):
                if x is None:
                    return None

                flgs = 0
                if not case:
                    flgs = re.I
                if flags > 0:
                    flgs = flgs | flags
                r = re.compile(pat, flgs)
                return r.search(x) is not None

            func._identifier = "str_contains"
            pat = expr._pat if not isinstance(expr._pat, StringScalar) or expr._pat._value is None \
                else Scalar(re.escape(to_text(expr.pat)))
            inputs = expr.input, pat, expr._case, expr._flags
            sub = self._gen_mapped_expr(expr, inputs, func,
                                        expr.name, multiple=False)
            self._sub(expr, sub)
            return
        elif isinstance(expr, StrCount):
            def func(x, pat, flags):
                if x is None:
                    return None

                regex = re.compile(pat, flags=flags)
                return len(regex.findall(x))

            func._identifier = "str_count"
            pat = expr._pat if not isinstance(expr._pat, StringScalar) or expr._pat._value is None \
                else Scalar(re.escape(to_text(expr.pat)))
            inputs = expr.input, pat, expr._flags
            sub = self._gen_mapped_expr(expr, inputs, func, expr.name, multiple=False)
            self._sub(expr, sub)
            return
        elif isinstance(expr, Find) and expr.end is not None:
            start = expr.start
            end = expr.end
            substr = expr.sub

            def func(x):
                if x is None:
                    return None

                return x.find(substr, start, end)

            func._identifier = "str_find"
        elif isinstance(expr, RFind):
            start = expr.start
            end = expr.end
            substr = expr.sub

            def func(x):
                if x is None:
                    return None

                return x.rfind(substr, start, end)

            func._identifier = "str_rfind"
        elif isinstance(expr, Extract):
            def func(x, pat, flags, group):
                if x is None:
                    return None

                regex = re.compile(pat, flags=flags)
                m = regex.search(x)
                if m:
                    if group is None:
                        return m.group()
                    return m.group(group)

            func._identifier = "str_extract"
            pat = expr._pat if not isinstance(expr._pat, StringScalar) or expr._pat._value is None \
                else Scalar(re.escape(to_text(expr.pat)))
            inputs = expr.input, pat, expr._flags, expr._group
            sub = self._gen_mapped_expr(expr, inputs, func, expr.name, multiple=False)
            self._sub(expr, sub)
            return
        elif isinstance(expr, Replace):
            def func(x, pat, repl, n, case, flags, use_regex):
                if x is None:
                    return None

                use_re = use_regex and (not case or len(pat) > 1 or flags)

                if use_re:
                    if not case:
                        flags |= re.IGNORECASE
                    regex = re.compile(pat, flags=flags)
                    n = n if n >= 0 else 0

                    return regex.sub(repl, x, count=n)
                else:
                    return x.replace(pat, repl, n)

            func._identifier = "str_replace"
            pat = expr._pat if not isinstance(expr._pat, StringScalar) or expr._pat._value is None \
                else Scalar(re.escape(to_text(expr.pat)))
            inputs = expr.input, pat, expr._repl, expr._n, \
                     expr._case, expr._flags, expr._regex
            sub = self._gen_mapped_expr(expr, inputs, func,
                                        expr.name, multiple=False)
            self._sub(expr, sub)
            return
        elif isinstance(expr, (Lstrip, Strip, Rstrip)) and expr.to_strip != ' ':
            to_strip = expr.to_strip

            if isinstance(expr, Lstrip):
                def func(x):
                    if x is None:
                        return None

                    return x.lstrip(to_strip)

                func._identifier = "str_lstrip"
            elif isinstance(expr, Strip):
                def func(x):
                    if x is None:
                        return None

                    return x.strip(to_strip)

                func._identifier = "str_strip"
            elif isinstance(expr, Rstrip):
                def func(x):
                    if x is None:
                        return None

                    return x.rstrip(to_strip)

                func._identifier = "str_rstrip"
        elif isinstance(expr, Pad):
            side = expr.side
            fillchar = expr.fillchar
            width = expr.width

            def func(x, width, fillchar, side):
                if x is None:
                    return None
                if side == 'left':
                    return x.rjust(width, fillchar)
                elif side == 'right':
                    return x.ljust(width, fillchar)
                else:
                    return x.center(width, fillchar)

            func._identifier = "str_pad"
            if side not in ('left', 'right', 'both'):
                raise NotImplementedError
            inputs = expr.input, Scalar(width), Scalar(fillchar), Scalar(side)
            sub = self._gen_mapped_expr(expr, inputs, func, expr.name, multiple=False)
            self._sub(expr, sub)
            return
        elif isinstance(expr, Slice):
            start, end, step = expr.start, expr.end, expr.step

            if end is None and step is None:
                raise NotImplementedError
            if isinstance(start, six.integer_types) and \
                    isinstance(end, six.integer_types) and step is None:
                if start >= 0 and end >= 0:
                    raise NotImplementedError

            flag = 0x4 if start is not None else 0
            flag |= 0x2 if end is not None else 0
            flag |= 0x1 if step is not None else 0

            def func(x, flag, *args):
                if x is None:
                    return None

                idx = 0
                s, e, t = None, None, None
                for i in range(3):
                    if i == 0 and (flag & 0x4):
                        s = args[idx]
                        idx += 1
                    if i == 1 and (flag & 0x2):
                        e = args[idx]
                        idx += 1
                    if i == 2 and (flag & 0x1):
                        t = args[idx]
                        idx += 1
                return x[s: e: t]

            func._identifier = "str_slice"
            inputs = expr.input, Scalar(flag), expr._start, expr._end, expr._step
            sub = self._gen_mapped_expr(expr, tuple(i for i in inputs if i is not None),
                                        func, expr.name, multiple=False)
            self._sub(expr, sub)
            return
        elif isinstance(expr, Swapcase):
            func = lambda x: x.swapcase() if x is not None else None
            func._identifier = "str_swapcase"
        elif isinstance(expr, Title):
            func = lambda x: x.title() if x is not None else None
            func._identifier = "str_title"
        elif isinstance(expr, Strptime):
            def func(x, date_fmt):
                from datetime import datetime

                return datetime.strptime(x, date_fmt) if x is not None else None

            func._identifier = "strptime"
            date_fmt = expr._date_format \
                if not isinstance(expr._date_format, StringScalar) or expr._date_format._value is None \
                else Scalar(to_text(expr.date_format).replace("%", "%%"))
            inputs = expr.input, date_fmt
            sub = self._gen_mapped_expr(expr, inputs, func, expr.name, multiple=False)
            self._sub(expr, sub)
            return
        else:
            if isinstance(expr, Isalnum):
                func = lambda x: x.isalnum() if x is not None else None
                func._identifier = "str_isalnum"
            elif isinstance(expr, Isalpha):
                func = lambda x: x.isalpha() if x is not None else None
                func._identifier = "str_isalpha"
            elif isinstance(expr, Isdigit):
                func = lambda x: x.isdigit() if x is not None else None
                func._identifier = "str_isdigit"
            elif isinstance(expr, Isspace):
                func = lambda x: x.isspace() if x is not None else None
                func._identifier = "str_isspace"
            elif isinstance(expr, Islower):
                func = lambda x: x.islower() if x is not None else None
                func._identifier = "str_islower"
            elif isinstance(expr, Isupper):
                func = lambda x: x.isupper() if x is not None else None
                func._identifier = "str_isupper"
            elif isinstance(expr, Istitle):
                func = lambda x: x.istitle() if x is not None else None
                func._identifier = "str_istitle"
            elif isinstance(expr, (Isnumeric, Isdecimal)):
                def u_safe(s):
                    try:
                        return unicode(s, "unicode_escape")
                    except:
                        return s

                if isinstance(expr, Isnumeric):
                    func = lambda x: u_safe(x).isnumeric() if x is not None else None
                    func._identifier = "str_isnumeric"
                else:
                    func = lambda x: u_safe(x).isdecimal() if x is not None else None
                    func._identifier = "str_isdecimal"

        if func is not None:
            sub = expr.input.map(func, expr.dtype)
            self._sub(expr, sub)
            return

        raise NotImplementedError

    def visit_rank_window(self, expr):
        if isinstance(expr, QCut):
            self._sub(expr, expr - 1)
            return

        raise NotImplementedError

    def visit_reduction(self, expr):
        expr_input = expr.input
        if getattr(expr, '_unique', False):
            expr_input = expr_input.unique()

        if isinstance(expr, (Var, GroupedVar)):
            std = expr_input.std(ddof=expr._ddof)
            if isinstance(expr, GroupedVar):
                std = std.to_grouped_reduction(expr._grouped)
            sub = (std ** 2).rename(expr.name)
            self._sub(expr, sub)
            return
        elif isinstance(expr, (Moment, GroupedMoment)):
            order = expr._order
            center = expr._center

            sub = self._get_moment_sub_expr(expr, expr_input, order, center)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)
            return
        elif isinstance(expr, (Skewness, GroupedSkewness)):
            std = expr_input.std(ddof=1)
            if isinstance(expr, GroupedSequenceReduction):
                std = std.to_grouped_reduction(expr._grouped)
            cnt = expr_input.count()
            if isinstance(expr, GroupedSequenceReduction):
                cnt = cnt.to_grouped_reduction(expr._grouped)
            sub = self._get_moment_sub_expr(expr, expr_input, 3, True) / (std ** 3)
            sub *= (cnt ** 2) / (cnt - 1) / (cnt - 2)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)
        elif isinstance(expr, (Kurtosis, GroupedKurtosis)):
            std = expr_input.std(ddof=0)
            if isinstance(expr, GroupedSequenceReduction):
                std = std.to_grouped_reduction(expr._grouped)
            cnt = expr_input.count()
            if isinstance(expr, GroupedSequenceReduction):
                cnt = cnt.to_grouped_reduction(expr._grouped)
            m4 = self._get_moment_sub_expr(expr, expr_input, 4, True)
            sub = 1.0 / (cnt - 2) / (cnt - 3) * ((cnt * cnt - 1) * m4 / (std ** 4) - 3 * (cnt - 1) ** 2)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)

        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from collections import OrderedDict
from datetime import datetime

from .core import Iterable
from .instance import Instance
from .session import SessionInstance
from .job import Job
from .. import serializers, errors, utils
from ..compat import six
from ..config import options


class BaseInstances(Iterable):
    def _get(self, name):
        return Instance(client=self._client, parent=self, name=name)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            instance = self._get(item)
        elif isinstance(item, Instance):
            instance = item
        else:
            return False

        try:
            instance.reload()
            return True
        except errors.NoSuchObject:
            return False

    def __iter__(self):
        return self.iterate()

    def iterate(self, start_time=None, end_time=None, status=None, only_owner=None,
                max_items=None, job_name=None, quota_index=None, **kw):
        if 'from_time' in kw:
            start_time = kw['from_time']

        if isinstance(status, six.string_types):
            status = Instance.Status(status.capitalize())

        params = dict()
        if status is not None:
            params['status'] = status.value
        if start_time is not None or end_time is not None:
            daterange = six.StringIO()
            if start_time is not None:
                if isinstance(start_time, datetime):
                    daterange.write(str(utils.to_timestamp(start_time)))
                else:
                    daterange.write(str(int(start_time)))
            daterange.write(':')
            if end_time is not None:
                if isinstance(end_time, datetime):
                    daterange.write(str(utils.to_timestamp(end_time)))
                else:
                    daterange.write(str(int(end_time)))
            params['daterange'] = daterange.getvalue()
        if only_owner is not None:
            params['onlyowner'] = 'yes' if only_owner else 'no'
        if max_items is not None:
            params['maxitems'] = max_items
        if job_name is not None:
            params['jobname'] = job_name
        if quota_index is not None:
            params['quotaIndex'] = quota_index

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                    (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, params=params)

            inst = Instances.parse(self._client, resp, obj=self)
            params['marker'] = inst.marker

            return inst.instances

        while True:
            instances = _it()
            if instances is None:
                break
            for instance in instances:
                yield instance

    @classmethod
    def _create_job(cls, job=None, task=None, priority=None, running_cluster=None, uuid_=None):
        job = job or Job()
        if priority is not None:
            if priority < 0:
                raise errors.ODPSClientError('Priority must more than or equal to zero.')
            job.priority = priority
        if running_cluster is not None:
            job.running_cluster = running_cluster
        if task is not None:
            job.add_task(task)
        if job.tasks is None or len(job.tasks) == 0:
            raise ValueError('Job tasks are required')

        guid = uuid_ or str(uuid.uuid4())
        for t in job.tasks:
            cls._fill_task_properties(t, uuid_=guid)
        return job

    @classmethod
    def _fill_task_properties(cls, task, uuid_=None):
        if task.properties is None:
            task.properties = OrderedDict()
        if options.biz_id:
            task.properties['biz_id'] = str(options.biz_id)

        guid = uuid_ or str(uuid.uuid4())
        if task.properties.get('uuid') is None:
            task.set_property('uuid', guid)
        if task.name is None:
            raise errors.ODPSClientError('Task name is required')

    @classmethod
    def _get_submit_instance_content(cls, job):
        return Instance.AnonymousSubmitInstance(job=job).serialize()

    def create(self, xml=None, job=None, task=None, priority=None, running_cluster=None,
               headers=None, create_callback=None, encoding=None, session_project=None,
               session_name=None):
        if xml is None:
            job = self._create_job(
                job=job, task=task, priority=priority, running_cluster=running_cluster
            )

            xml = self._get_submit_instance_content(job)

        headers = headers or dict()
        headers['Content-Type'] = 'application/xml'
        url = self.resource()
        resp = self._client.post(url, xml, headers=headers)

        location = resp.headers.get('Location')
        if location is None or len(location) == 0:
            raise errors.ODPSClientError(
                'Invalid response, Location header required. As it is a rare '
                'condition, please check your network policies first.',
                request_id=resp.headers.get('x-odps-request-id'),
            )

        instance_id = location.rsplit('/', 1)[1]

        create_callback = create_callback or options.instance_create_callback
        if create_callback is not None:
            create_callback(instance_id)

        if encoding is not None:
            resp.encoding = encoding
        body = resp.text
        if body:
            instance_result = Instance.InstanceResult.parse(self._client, resp)
            results = dict([(r.name, r.result) for r in instance_result.task_results])
        else:
            results = None

        if session_project:
            instance = SessionInstance(session_project=session_project,
                                       session_task_name=task.name,
                                       session_name=session_name,
                                       name=instance_id, task_results=results,
                                       parent=self, client=self._client)
        else:
            instance = Instance(name=instance_id, task_results=results,
                                parent=self, client=self._client)
        return instance


class Instances(BaseInstances):

    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems')
    instances = serializers.XMLNodesReferencesField(Instance, 'Instance')


class CachedInstances(BaseInstances):

    class _CachedInstances(serializers.JSONSerializableModel):
        instance_queueing_infos = \
            serializers.JSONNodesReferencesField(Instance.InstanceQueueingInfo)

    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems')
    instances = serializers.XMLNodeReferenceField(_CachedInstances, 'Content')

    def iterate(self, status=None, only_owner=None,
                max_items=None, quota_index=None, **kw):
        if isinstance(status, six.string_types):
            status = Instance.Status(status.capitalize())

        params = dict()
        if status is not None:
            params['status'] = status.value
        if only_owner is not None:
            params['onlyowner'] = 'yes' if only_owner else 'no'
        if max_items is not None:
            params['maxitems'] = max_items
        if quota_index is not None:
            params['quotaIndex'] = quota_index

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, params=params)

            inst = CachedInstances.parse(self._client, resp, obj=self)
            params['marker'] = inst.marker

            return inst.instances.instance_queueing_infos

        while True:
            instances = _it()
            if instances is None:
                break
            for instance in instances:
                yield instance



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
from ctypes import CDLL, RTLD_GLOBAL, create_string_buffer, \
    c_char, c_void_p, c_char_p, c_int32, c_uint32, byref, POINTER
from ctypes.util import find_library
from io import IOBase

from odps.compat import BytesIO, six
from odps.utils import to_binary
from ..errors import SubprocessStreamEOFError, CupidCppError
from ..utils import get_environ

_lib_path = find_library('odps_subprocess')
if _lib_path is None:
    ld_paths = (get_environ('LD_LIBRARY_PATH') or "").split(':')
    for p in ld_paths:
        so_path = os.path.join(p, 'libodps_subprocess.so')
        if os.path.exists(so_path):
            _lib_path = so_path
            break
if _lib_path is None:
    _lib_path = os.path.join(os.getcwd(), 'libodps_subprocess.so')

try:
    odps_subproc = CDLL(_lib_path, mode=RTLD_GLOBAL)
except OSError:
    raise ImportError('Failed to load libodps_subprocess.so.')

Subprocess_Container_Init = odps_subproc.Subprocess_Container_Init

Subprocess_StdString_Size = odps_subproc.Subprocess_StdString_Size
Subprocess_StdString_Size.argtypes = [c_void_p]
Subprocess_StdString_Size.restype = c_uint32

Subprocess_StdString_CopyTo = odps_subproc.Subprocess_StdString_CopyTo
Subprocess_StdString_CopyTo.argtypes = [c_char_p, c_void_p, c_uint32]

Subprocess_StdString_delete = odps_subproc.Subprocess_StdString_delete
Subprocess_StdString_delete.argtypes = [c_void_p]

Subprocess_StartFDReceiver = odps_subproc.Subprocess_StartFDReceiver

Subprocess_StdIStream_eof = odps_subproc.Subprocess_StdIStream_eof
Subprocess_StdIStream_eof.argtypes = [c_void_p]
Subprocess_StdIStream_eof.restype = c_int32

Subprocess_StdIStream_read = odps_subproc.Subprocess_StdIStream_read
Subprocess_StdIStream_read.argtypes = [c_void_p, c_char_p, c_uint32]
Subprocess_StdIStream_read.restype = c_uint32

Subprocess_StdIStream_seekg = odps_subproc.Subprocess_StdIStream_seekg
Subprocess_StdIStream_seekg.argtypes = [c_void_p, c_int32, c_int32]

Subprocess_StdIStream_tellg = odps_subproc.Subprocess_StdIStream_tellg
Subprocess_StdIStream_tellg.argtypes = [c_void_p]
Subprocess_StdIStream_tellg.restype = c_uint32

Subprocess_StdIStream_getline = odps_subproc.Subprocess_StdIStream_getline
Subprocess_StdIStream_getline.argtypes = [c_void_p]
Subprocess_StdIStream_getline.restype = c_void_p

Subprocess_StdIStream_delete = odps_subproc.Subprocess_StdIStream_delete
Subprocess_StdIStream_delete.argtypes = [c_void_p]

Subprocess_StdOStream_write = odps_subproc.Subprocess_StdOStream_write
Subprocess_StdOStream_write.argtypes = [c_void_p, c_char_p, c_uint32]

Subprocess_StdOStream_flush = odps_subproc.Subprocess_StdOStream_flush
Subprocess_StdOStream_flush.argtypes = [c_void_p]

Subprocess_StdOStream_delete = odps_subproc.Subprocess_StdOStream_delete
Subprocess_StdOStream_delete.argtypes = [c_void_p]

ChannelConf_new = odps_subproc.ChannelConf_new
ChannelConf_new.restype = c_void_p

ChannelConf_delete = odps_subproc.ChannelConf_delete
ChannelConf_delete.argtypes = [c_void_p]

ChannelConf_SetInt = odps_subproc.ChannelConf_SetInt
ChannelConf_SetInt.argtypes = [c_void_p, c_char_p, c_int32]

ChannelResultInputStreamPtr_GetResult = odps_subproc.ChannelResultInputStreamPtr_GetResult
ChannelResultInputStreamPtr_GetResult.argtypes = [c_void_p, c_int32, POINTER(c_void_p)]
ChannelResultInputStreamPtr_GetResult.restype = c_void_p

ChannelResultInputStreamPtr_GetStream = odps_subproc.ChannelResultInputStreamPtr_GetStream
ChannelResultInputStreamPtr_GetStream.argtypes = [c_void_p]
ChannelResultInputStreamPtr_GetStream.restype = c_void_p

ChannelResultInputStreamPtr_GetFD = odps_subproc.ChannelResultInputStreamPtr_GetFD
ChannelResultInputStreamPtr_GetFD.argtypes = [c_void_p]
ChannelResultInputStreamPtr_GetFD.restype = c_int32

ChannelResultInputStreamPtr_delete = odps_subproc.ChannelResultInputStreamPtr_delete
ChannelResultInputStreamPtr_delete.argtypes = [c_void_p]

ChannelResultOutputStreamPtr_GetResult = odps_subproc.ChannelResultOutputStreamPtr_GetResult
ChannelResultOutputStreamPtr_GetResult.argtypes = [c_void_p, c_int32, POINTER(c_void_p)]
ChannelResultOutputStreamPtr_GetResult.restype = c_void_p

ChannelResultOutputStreamPtr_GetStream = odps_subproc.ChannelResultOutputStreamPtr_GetStream
ChannelResultOutputStreamPtr_GetStream.argtypes = [c_void_p]
ChannelResultOutputStreamPtr_GetStream.restype = c_void_p

ChannelResultOutputStreamPtr_GetFD = odps_subproc.ChannelResultOutputStreamPtr_GetFD
ChannelResultOutputStreamPtr_GetFD.argtypes = [c_void_p]
ChannelResultOutputStreamPtr_GetFD.restype = c_int32

ChannelResultOutputStreamPtr_delete = odps_subproc.ChannelResultOutputStreamPtr_delete
ChannelResultOutputStreamPtr_delete.argtypes = [c_void_p]

ChannelSlaveClient_new = odps_subproc.ChannelSlaveClient_new
ChannelSlaveClient_new.argtypes = [c_int32, c_int32, c_char_p]
ChannelSlaveClient_new.restype = c_void_p

ChannelSlaveClient_delete = odps_subproc.ChannelSlaveClient_delete
ChannelSlaveClient_delete.argtypes = [c_void_p]

ChannelSlaveClient_Start = odps_subproc.ChannelSlaveClient_Start
ChannelSlaveClient_Start.argtypes = [c_void_p]

ChannelSlaveClient_Stop = odps_subproc.ChannelSlaveClient_Stop
ChannelSlaveClient_Stop.argtypes = [c_void_p]

ChannelSlaveClient_SyncCall = odps_subproc.ChannelSlaveClient_SyncCall
ChannelSlaveClient_SyncCall.argtypes = [c_void_p, c_char_p, c_char_p, c_uint32, c_int32, POINTER(c_void_p)]
ChannelSlaveClient_SyncCall.restype = c_void_p

ChannelSlaveClient_CreateInputStream = odps_subproc.ChannelSlaveClient_CreateInputStream
ChannelSlaveClient_CreateInputStream.argtypes = [c_void_p, c_char_p, c_char_p, c_uint32, POINTER(c_void_p)]
ChannelSlaveClient_CreateInputStream.restype = c_void_p

ChannelSlaveClient_CreateOutputStream = odps_subproc.ChannelSlaveClient_CreateOutputStream
ChannelSlaveClient_CreateOutputStream.argtypes = [c_void_p, c_char_p, c_char_p, c_uint32, POINTER(c_void_p)]
ChannelSlaveClient_CreateOutputStream.restype = c_void_p

ChannelApiError_GetErrorType = odps_subproc.ChannelApiError_GetErrorType
ChannelApiError_GetErrorType.argtypes = [c_void_p]
ChannelApiError_GetErrorType.restype = c_void_p

ChannelApiError_GetErrorMessage = odps_subproc.ChannelApiError_GetErrorMessage
ChannelApiError_GetErrorMessage.argtypes = [c_void_p]
ChannelApiError_GetErrorMessage.restype = c_void_p

ChannelApiError_delete = odps_subproc.ChannelApiError_delete
ChannelApiError_delete.argtypes = [c_void_p]

READ_CHUNK = 2048


def _read_std_string_ptr(std_string):
    std_string_len = Subprocess_StdString_Size(std_string)
    p = create_string_buffer(std_string_len)
    Subprocess_StdString_CopyTo(p, std_string, std_string_len)
    Subprocess_StdString_delete(std_string)
    return p


def _call_with_raise(func, *args):
    err_pp = c_void_p()
    err_ptr = 0
    try:
        args = args + (byref(err_pp), )
        res = func(*args)
        if res:
            return res
        else:
            err_ptr = err_pp.value
            err_type_ptr = ChannelApiError_GetErrorType(err_ptr)
            err_type = _read_std_string_ptr(err_type_ptr).raw
            err_message_ptr = ChannelApiError_GetErrorMessage(err_ptr)
            err_message = _read_std_string_ptr(err_message_ptr).raw
            raise CupidCppError(err_type, err_message)
    finally:
        if err_ptr:
            ChannelApiError_delete(err_ptr)


class IStreamWrapper(IOBase):
    def __init__(self, ptr, free=True):
        self._ptr = ptr
        self._free = free

    def __del__(self):
        self.close()

    def close(self):
        if self._free and self._ptr is not None:
            Subprocess_StdIStream_delete(self._ptr)
            self._ptr = None

    @property
    def closed(self):
        if self._ptr is None:
            return True
        else:
            return False

    def _reset(self):
        self._ptr = None

    def readable(self):
        return self._ptr is not None

    def read(self, size=-1):
        if Subprocess_StdIStream_eof(self._ptr):
            raise SubprocessStreamEOFError('Subprocess stream exhausted')
        if size < 0:
            buf_io = BytesIO()
            while True:
                try:
                    chunk = self.read(READ_CHUNK)
                except SubprocessStreamEOFError:
                    chunk = b''
                buf_io.write(chunk)
                if len(chunk) < READ_CHUNK:
                    break
            return buf_io.getvalue()
        else:
            p = create_string_buffer(size)
            Subprocess_StdIStream_read(self._ptr, p, size)
            return p.raw

    def readinto(self, b, offset=0):
        if Subprocess_StdIStream_eof(self._ptr):
            raise SubprocessStreamEOFError('Subprocess stream exhausted')
        size = len(b) - offset
        array_cls = c_char * size
        p = array_cls.from_buffer(b, offset)
        ret_size = Subprocess_StdIStream_read(self._ptr, p, size)
        if ret_size == 0 and Subprocess_StdIStream_eof(self._ptr):
            raise SubprocessStreamEOFError('Subprocess stream exhausted')
        return ret_size

    def readline(self, size=-1):
        str_ptr = Subprocess_StdIStream_getline(self._ptr, size)
        p = _read_std_string_ptr(str_ptr)
        return p.raw

    def readlines(self, hint=-1):
        lines = []
        while True:
            line = self.readline(hint)
            lines.append(line)
            if hint == 0 or Subprocess_StdIStream_eof(self._ptr):
                break
            if hint > 0:
                hint -= len(line)
        return lines

    def __iter__(self):
        while True:
            line = self.readline()
            yield line
            if Subprocess_StdIStream_eof(self._ptr):
                break

    def seek(self, offset, whence=0):
        Subprocess_StdIStream_seekg(self._ptr, offset, whence)

    def tell(self):
        return Subprocess_StdIStream_tellg(self._ptr)


class OStreamWrapper(IOBase):
    def __init__(self, ptr, free=True):
        self._ptr = ptr
        self._free = free

    def __del__(self):
        if self._ptr is not None:
            self.flush()
        self.close()

    def close(self):
        if self._free and self._ptr is not None:
            Subprocess_StdOStream_delete(self._ptr)
            self._ptr = None

    @property
    def closed(self):
        if self._ptr is None:
            return True
        else:
            return False

    def _reset(self):
        self._ptr = None

    def writable(self):
        return self._ptr is not None

    def write(self, data, length=None):
        length = length or len(data)
        atype = c_char * length
        Subprocess_StdOStream_write(self._ptr, atype.from_buffer_copy(data), length)

    def flush(self):
        if self._ptr is not None:
            Subprocess_StdOStream_flush(self._ptr)


class ChannelResultInputStreamWrapper(IStreamWrapper):
    def __init__(self, ptr):
        self._result_ptr = ptr
        is_ptr = ChannelResultInputStreamPtr_GetStream(ptr)
        super(ChannelResultInputStreamWrapper, self).__init__(is_ptr, False)

    def __del__(self):
        self.close()

    def fileno(self):
        return ChannelResultInputStreamPtr_GetFD(self._result_ptr)

    def result(self, timeout=-1):
        # Stream object already freed in C end
        self._reset()

        pstr = _call_with_raise(ChannelResultInputStreamPtr_GetResult, self._result_ptr, timeout)
        p = _read_std_string_ptr(pstr)
        return p.raw

    def close(self):
        if self._result_ptr is not None:
            ChannelResultInputStreamPtr_delete(self._result_ptr)
        self._result_ptr = None
        super(ChannelResultInputStreamWrapper, self).close()


class ChannelResultOutputStreamWrapper(OStreamWrapper):
    def __init__(self, ptr):
        self._result_ptr = ptr
        os_ptr = ChannelResultOutputStreamPtr_GetStream(ptr)
        super(ChannelResultOutputStreamWrapper, self).__init__(os_ptr, False)

    def __del__(self):
        self.close()

    def fileno(self):
        return ChannelResultOutputStreamPtr_GetFD(self._result_ptr)

    def result(self, timeout=-1):
        # Stream object already freed in C end
        self._reset()

        pstr = _call_with_raise(ChannelResultOutputStreamPtr_GetResult, self._result_ptr, timeout)
        p = _read_std_string_ptr(pstr)
        return p.raw

    def close(self):
        if self._result_ptr is not None:
            self.flush()
            ChannelResultOutputStreamPtr_delete(self._result_ptr)
        self._result_ptr = None
        super(ChannelResultOutputStreamWrapper, self).close()


class ChannelOutputWriter(object):
    def __init__(self, stream):
        self._stream = stream

    def write(self, data):
        self._stream.write(data)

    def close(self):
        # sync
        self._stream.result()
        self._stream.close()

    @property
    def closed(self):
        return self._stream.closed

    def flush(self):
        self._stream.flush()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


class ChannelConf(object):
    def __init__(self):
        self._ptr = ChannelConf_new()

    def __del__(self):
        if self._ptr is not None:
            ChannelConf_delete(self._ptr)
        self._ptr = None

    def set_integer(self, key, value):
        if isinstance(key, six.string_types):
            key = key.encode()
        ChannelConf_SetInt(self._ptr, key, value)


class ChannelSlaveClient(object):
    def __init__(self, request_channel_nums, response_channel_nums, client_name):
        if isinstance(client_name, six.string_types):
            client_name = client_name.encode()
        self._ptr = ChannelSlaveClient_new(request_channel_nums, response_channel_nums, client_name)

    def __del__(self):
        if self._ptr is not None:
            ChannelSlaveClient_delete(self._ptr)
        self._ptr = None

    def start(self):
        ChannelSlaveClient_Start(self._ptr)

    def stop(self):
        ChannelSlaveClient_Stop(self._ptr)

    def sync_call(self, method, parameter, timeout=-1):
        method = to_binary(method)
        parameter = to_binary(parameter)
        result_ptr = _call_with_raise(ChannelSlaveClient_SyncCall, self._ptr, method,
                                      parameter, len(parameter), timeout)
        p = _read_std_string_ptr(result_ptr)
        return p.raw

    def create_file_reader(self, method, params):
        if isinstance(method, six.string_types):
            method = method.encode()
        if isinstance(params, six.string_types):
            params = params.encode()
        ptr = _call_with_raise(ChannelSlaveClient_CreateInputStream, self._ptr, method, params, len(params))
        return ChannelResultInputStreamWrapper(ptr)

    def create_file_writer(self, method, params):
        if isinstance(method, six.string_types):
            method = method.encode()
        if isinstance(params, six.string_types):
            params = params.encode()
        ptr = _call_with_raise(ChannelSlaveClient_CreateOutputStream, self._ptr, method, params, len(params))
        return ChannelResultOutputStreamWrapper(ptr)

    def create_record_reader(self, label, schema, columns=None):
        from ..io.table import CupidRecordReader

        params = dict(type='ReadByLabel', label=label, arrow=False, batch=False)
        stream = self.create_file_reader('createTableInputStream', json.dumps(params).encode())
        reader = CupidRecordReader(schema, stream, columns=columns)
        return reader

    def create_record_writer(self, label, schema):
        from ..io.table import CupidRecordWriter
        params = json.dumps(dict(type='WriteByLabel', label=label, arrow=False, Batch=False))
        stream = self.create_file_writer('createTableOutputStream', params.encode())
        writer = CupidRecordWriter(schema, stream)
        return writer

    def create_arrow_writer(self, label):
        params = json.dumps(dict(type='WriteByLabel', label=label, arrow=True, batch=True))
        stream = self.create_file_writer('createTableOutputStream', params.encode())
        return ChannelOutputWriter(stream)

    create_table_reader = create_record_reader
    create_table_writer = create_record_writer

    def create_pandas_reader(self, label, schema, columns=None):
        try:
            from ..io.table import CupidPandasReader
        except ImportError:
            return None

        params = json.dumps(dict(type='ReadByLabel', label=label))
        stream = self.create_file_reader('createTableInputStream', params.encode())
        reader = CupidPandasReader(schema, stream, columns=columns)
        return reader

    def create_pandas_writer(self, label, schema):
        try:
            from ..io.table import CupidPandasWriter
        except ImportError:
            return None

        params = json.dumps(dict(type='WriteByLabel', label=label))
        stream = self.create_file_writer('createTableOutputStream', params.encode())
        writer = CupidPandasWriter(schema, stream)
        return writer
